"""
    Forum attachments model admin definitions
    =========================================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from django.contrib import admin

from machina.core.db.models import get_model


Attachment = get_model('forum_attachments', 'Attachment')


class AttachmentAdmin(admin.ModelAdmin):
    """ The Attachment model admin. """

    list_display = ('id', 'post', 'comment', 'file', )
    list_display_links = ('id', 'post', 'comment', )
    raw_id_fields = ('post', )


admin.site.register(Attachment, AttachmentAdmin)



"""
    Forum attachments URLs
    ======================

    This module defines URL patterns associated with the django-machina's ``forum_attachments``
    application.

"""

from django.urls import path

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumAttachmentsURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_attachments`` application. """

    attachment_view = get_class('forum_conversation.forum_attachments.views', 'AttachmentView')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path('attachment/<int:pk>/', self.attachment_view.as_view(), name='attachment'),
        ]


urlpatterns_factory = ForumAttachmentsURLPatternsFactory()



"""
    Forum polls URLs
    ================

    This module defines URL patterns associated with the django-machina's ``forum_polls``
    application.

"""

from django.urls import path

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumPollsURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_polls`` application. """

    poll_vote_view = get_class('forum_conversation.forum_polls.views', 'TopicPollVoteView')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path('poll/<int:pk>/vote/', self.poll_vote_view.as_view(), name='topic_poll_vote'),
        ]


urlpatterns_factory = ForumPollsURLPatternsFactory()



"""
    Forum model admin definitions
    =============================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from collections import OrderedDict

from django.conf.urls import url
from django.contrib import admin
from django.contrib.admin import helpers
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.forms.forms import NON_FIELD_ERRORS
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from mptt.exceptions import InvalidMove

from machina.core.db.models import get_model
from machina.core.loading import get_class
from machina.models.fields import MarkupTextField, MarkupTextFieldWidget


Forum = get_model('forum', 'Forum')
ForumPermission = get_model('forum_permission', 'ForumPermission')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')

PermissionsForm = get_class('forum.forms', 'PermissionsForm')
PickForumForm = get_class('forum.forms', 'PickForumForm')
PickGroupForm = get_class('forum.forms', 'PickGroupForm')
PickUserForm = get_class('forum.forms', 'PickUserForm')

PermissionConfig = get_class('forum_permission.defaults', 'PermissionConfig')


class ForumAdmin(admin.ModelAdmin):
    """ The Forum model admin. """

    formfield_overrides = {
        MarkupTextField: {'widget': MarkupTextFieldWidget},
    }

    fieldsets = (
        [None, {'fields': ('type', 'parent', 'name', 'description', 'image')}],
        [_('Forum settings'), {'fields': ('display_sub_forum_list', ), 'classes': ('collapse', )}],
        [
            _('Link forum settings'),
            {'fields': ('link', 'link_redirects', ), 'classes': ('collapse', )},
        ],
    )
    list_display = ('name', 'type', 'direct_topics_count', 'direct_posts_count', )
    search_fields = ('name', )

    editpermissions_index_view_template_name = 'admin/forum/forum/editpermissions_index.html'
    editpermissions_user_view_template_name = 'admin/forum/forum/editpermissions_user.html'
    editpermissions_anonymous_user_view_template_name = (
        'admin/forum/forum/editpermissions_anonymous_user.html'
    )
    editpermissions_authenticated_user_view_template_name = (
        'admin/forum/forum/editpermissions_authenticated_user.html'
    )
    editpermissions_group_view_template_name = 'admin/forum/forum/editpermissions_group.html'

    def get_urls(self):
        """ Returns the URLs associated with the admin abstraction. """
        urls = super().get_urls()
        forum_admin_urls = [
            url(
                r'^(?P<forum_id>[0-9]+)/move-forum/(?P<direction>up|down)/$',
                self.admin_site.admin_view(self.moveforum_view),
                name='forum_forum_move',
            ),
            url(
                r'^edit-global-permissions/$',
                self.admin_site.admin_view(self.editpermissions_index_view),
                name='forum_forum_editpermission_index',
            ),
            url(
                r'^edit-global-permissions/user/(?P<user_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_user_view),
                name='forum_forum_editpermission_user',
            ),
            url(
                r'^edit-global-permissions/user/anonymous/$',
                self.admin_site.admin_view(self.editpermissions_anonymous_user_view),
                name='forum_forum_editpermission_anonymous_user',
            ),
            url(
                r'^edit-global-permissions/user/authenticated/$',
                self.admin_site.admin_view(self.editpermissions_authenticated_user_view),
                name='forum_forum_editpermission_authenticated_user',
            ),
            url(
                r'^edit-global-permissions/group/(?P<group_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_group_view),
                name='forum_forum_editpermission_group',
            ),
            url(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/$',
                self.admin_site.admin_view(self.editpermissions_index_view),
                name='forum_forum_editpermission_index',
            ),
            url(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/user/(?P<user_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_user_view),
                name='forum_forum_editpermission_user',
            ),
            url(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/user/anonymous/$',
                self.admin_site.admin_view(self.editpermissions_anonymous_user_view),
                name='forum_forum_editpermission_anonymous_user',
            ),
            url(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/user/authenticated/$',
                self.admin_site.admin_view(self.editpermissions_authenticated_user_view),
                name='forum_forum_editpermission_authenticated_user',
            ),
            url(
                r'^(?P<forum_id>[0-9]+)/edit-permissions/group/(?P<group_id>[0-9]+)/$',
                self.admin_site.admin_view(self.editpermissions_group_view),
                name='forum_forum_editpermission_group',
            ),
        ]
        return forum_admin_urls + urls

    def get_forum_perms_base_context(self, request, obj=None):
        """ Returns the context to provide to the template for permissions contents. """
        context = {
            'adminform': {'model_admin': self},
            'media': self.media,
            'object': obj,
            'app_label': self.model._meta.app_label,
            'opts': self.model._meta,
            'has_change_permission': self.has_change_permission(request, obj),
        }
        try:
            context.update(self.admin_site.each_context(request))
        except TypeError:  # pragma: no cover
            # Django 1.7 compatibility
            context.update(self.admin_site.each_context())
        except AttributeError:  # pragma: no cover
            pass
        return context

    def moveforum_view(self, request, forum_id, direction):
        """ Moves the given forum toward the requested direction. """
        forum = get_object_or_404(Forum, pk=forum_id)

        # Fetch the target
        target, position = None, None
        if direction == 'up':
            target, position = forum.get_previous_sibling(), 'left'
        elif direction == 'down':
            target, position = forum.get_next_sibling(), 'right'

        # Do the move
        try:
            assert target is not None
            forum.move_to(target, position)
        except (InvalidMove, AssertionError):
            pass
        self.message_user(request, _("'{}' forum successfully moved").format(forum.name))
        return HttpResponseRedirect(reverse('admin:forum_forum_changelist'))

    def editpermissions_index_view(self, request, forum_id=None):
        """ Allows to select how to edit forum permissions.

        The view displays a form to select a user or a group in order to edit its permissions for
        the considered forum.

        """
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id \
            else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = _('Forum permissions') if forum else _('Global forum permissions')

        can_change_user_perms = (
            request.user.has_perm('forum_permission.add_userforumpermission') or
            request.user.has_perm('forum_permission.change_userforumpermission')
        )
        can_change_group_perms = (
            request.user.has_perm('forum_permission.add_groupforumpermission') or
            request.user.has_perm('forum_permission.change_groupforumpermission')
        )

        # Handles "copy permission from" form
        permissions_copied = False
        if can_change_user_perms and can_change_group_perms:
            if forum and request.method == 'POST':
                forum_form = PickForumForm(request.POST)
                if forum_form.is_valid() and forum_form.cleaned_data['forum']:
                    self._copy_forum_permissions(forum_form.cleaned_data['forum'], forum)
                    self.message_user(request, _('Permissions successfully copied'))
                    permissions_copied = True
                context['forum_form'] = forum_form
            elif forum:
                context['forum_form'] = PickForumForm()

        user_form, user = None, None
        group_form, group = None, None

        # Handles user or group selection
        if request.method == 'POST' and not permissions_copied:
            if can_change_user_perms:
                user_form = PickUserForm(request.POST, admin_site=self.admin_site)
            if can_change_group_perms:
                group_form = PickGroupForm(request.POST, admin_site=self.admin_site)

            # Check if the form was drawn, was valid and was submitted (presence of submitbutton)
            if user_form and user_form.is_valid() and '_select_user' in request.POST:
                user = user_form.cleaned_data.get('user', None) if user_form.cleaned_data else None
                anonymous_user = (
                    user_form.cleaned_data.get('anonymous_user', None)
                    if user_form.cleaned_data else None
                )
                authenticated_user = (
                    user_form.cleaned_data.get('authenticated_user', None)
                    if user_form.cleaned_data else None
                )

                if not user and not anonymous_user and not authenticated_user:
                    user_form._errors[NON_FIELD_ERRORS] = user_form.error_class([
                        _(
                            "Choose either a user ID, the anonymous user or the authenticated user"
                        ),
                    ])
                elif user:
                    # Redirect to user
                    url_kwargs = (
                        {'forum_id': forum.id, 'user_id': user.id}
                        if forum else {'user_id': user.id}
                    )
                    return redirect(
                        reverse('admin:forum_forum_editpermission_user', kwargs=url_kwargs),
                    )
                elif anonymous_user:
                    # Redirect to anonymous user
                    url_kwargs = {'forum_id': forum.id} if forum else {}
                    return redirect(
                        reverse(
                            'admin:forum_forum_editpermission_anonymous_user',
                            kwargs=url_kwargs,
                        ),
                    )
                elif authenticated_user:
                    # Redirect to authenticated user
                    url_kwargs = {'forum_id': forum.id} if forum else {}
                    return redirect(
                        reverse(
                            'admin:forum_forum_editpermission_authenticated_user',
                            kwargs=url_kwargs,
                        ),
                    )

                context['user_errors'] = helpers.AdminErrorList(user_form, [])

            # Check if the form was drawn, was valid and was submitted (presence of submitbutton)
            if group_form and group_form.is_valid() and '_select_group' in request.POST:
                group = group_form.cleaned_data.get('group', None) \
                    if group_form.cleaned_data else None

                if not group:
                    group_form._errors[NON_FIELD_ERRORS] = group_form.error_class(
                        [_('Choose a group ID'), ])
                else:
                    # Redirect to group
                    url_kwargs = (
                        {'forum_id': forum.id, 'group_id': group.id}
                        if forum else {'group_id': group.id}
                    )
                    return redirect(
                        reverse('admin:forum_forum_editpermission_group', kwargs=url_kwargs),
                    )

                context['group_errors'] = helpers.AdminErrorList(group_form, [])
        else:
            if can_change_user_perms:
                user_form = PickUserForm(admin_site=self.admin_site)
            if can_change_group_perms:
                group_form = PickGroupForm(admin_site=self.admin_site)

        context['user_form'] = user_form
        context['group_form'] = group_form

        return render(request, self.editpermissions_index_view_template_name, context)

    def editpermissions_user_view(self, request, user_id, forum_id=None):
        """ Allows to edit user permissions for the considered forum.

        The view displays a form to define which permissions are granted for the given user for the
        considered forum.

        """
        user_model = get_user_model()
        user = get_object_or_404(user_model, pk=user_id)
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), user)
        context['form'] = self._get_permissions_form(
            request, UserForumPermission, {'forum': forum, 'user': user},
        )

        return render(request, self.editpermissions_user_view_template_name, context)

    def editpermissions_anonymous_user_view(self, request, forum_id=None):
        """ Allows to edit anonymous user permissions for the considered forum.

        The view displays a form to define which permissions are granted for the anonymous user for
        the considered forum.

        """
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), _('Anonymous user'))
        context['form'] = self._get_permissions_form(
            request, UserForumPermission, {'forum': forum, 'anonymous_user': True},
        )

        return render(request, self.editpermissions_anonymous_user_view_template_name, context)

    def editpermissions_authenticated_user_view(self, request, forum_id=None):
        """ Allows to edit authenticated user permissions for the considered forum.

        The view displays a form to define which permissions are granted for the authenticated,
        non-specific, user for the considered forum.

        """
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), _('Authenticated user'))
        context['form'] = self._get_permissions_form(
            request, UserForumPermission, {'forum': forum, 'authenticated_user': True},
        )

        return render(request, self.editpermissions_authenticated_user_view_template_name, context)

    def editpermissions_group_view(self, request, group_id, forum_id=None):
        """ Allows to edit group permissions for the considered forum.

        The view displays a form to define which permissions are granted for the given group for the
        considered forum.

        """
        group = get_object_or_404(Group, pk=group_id)
        forum = get_object_or_404(Forum, pk=forum_id) if forum_id else None

        # Set up the context
        context = self.get_forum_perms_base_context(request, forum)
        context['forum'] = forum
        context['title'] = '{} - {}'.format(_('Forum permissions'), group)
        context['form'] = self._get_permissions_form(
            request, GroupForumPermission, {'forum': forum, 'group': group},
        )

        return render(request, self.editpermissions_group_view_template_name, context)

    def _get_permissions_form(self, request, permission_model, filter_kwargs):
        # Fetch the permissions
        editable_permissions = sorted(
            ForumPermission.objects.all(), key=lambda p: p.name,
        )
        granted_permissions = (
            permission_model.objects.filter(
                permission__in=editable_permissions, has_perm=True, **filter_kwargs
            )
            .values_list('permission__codename', flat=True)
        )
        non_granted_permissions = (
            permission_model.objects.filter(
                permission__in=editable_permissions, has_perm=False, **filter_kwargs
            )
            .values_list('permission__codename', flat=True)
        )

        permissions_dict = OrderedDict()
        for p in editable_permissions:
            if p.codename in granted_permissions:
                perm_state = PermissionsForm.PERM_GRANTED
            elif p.codename in non_granted_permissions:
                perm_state = PermissionsForm.PERM_NOT_GRANTED
            else:
                perm_state = PermissionsForm.PERM_NOT_SET
            permissions_dict[p.codename] = (p, perm_state)

        if request.method == 'POST':
            form = PermissionsForm(request.POST, permissions_dict=permissions_dict)
            if form.is_valid():
                for codename, value in form.cleaned_data.items():
                    try:
                        perm = permission_model.objects.get(
                            permission=permissions_dict[codename][0], **filter_kwargs
                        )
                    except permission_model.DoesNotExist:
                        if value == PermissionsForm.PERM_NOT_SET:
                            continue
                        perm = permission_model.objects.create(
                            permission=permissions_dict[codename][0], **filter_kwargs
                        )

                    if value == PermissionsForm.PERM_NOT_SET:
                        perm.delete()
                        continue

                    perm.has_perm = (value == PermissionsForm.PERM_GRANTED)
                    perm.save()

                self.message_user(request, _('Permissions successfully applied'))
        else:
            form = PermissionsForm(permissions_dict=permissions_dict)

        return form

    def _copy_forum_permissions(self, forum_from, forum_to):
        user_perms = UserForumPermission.objects.filter(forum=forum_from)
        group_perms = GroupForumPermission.objects.filter(forum=forum_from)

        for perm in user_perms:
            try:
                new_perm = UserForumPermission.objects.get(
                    permission=perm.permission, forum=forum_to,
                    user=perm.user, anonymous_user=perm.anonymous_user,
                )
            except UserForumPermission.DoesNotExist:
                new_perm = UserForumPermission(
                    permission=perm.permission, forum=forum_to,
                    user=perm.user, anonymous_user=perm.anonymous_user,
                    authenticated_user=perm.authenticated_user
                )
            new_perm.has_perm = perm.has_perm
            new_perm.save()

        for perm in group_perms:
            try:
                new_perm = GroupForumPermission.objects.get(
                    permission=perm.permission, forum=forum_to,
                    group=perm.group,
                )
            except GroupForumPermission.DoesNotExist:
                new_perm = GroupForumPermission(
                    permission=perm.permission, forum=forum_to,
                    group=perm.group,
                )
            new_perm.has_perm = perm.has_perm
            new_perm.save()


admin.site.register(Forum, ForumAdmin)



# Generated by Django 2.2.4 on 2019-08-21 08:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_permission', '0004_auto_20190319_2240'),
    ]

    operations = [
        migrations.AddField(
            model_name='userforumpermission',
            name='authenticated_user',
            field=models.BooleanField(db_index=True, default=False, verbose_name='Target any logged in user'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.3 on 2016-04-27 03:02
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0003_auto_20160228_2051'),
    ]

    operations = [
        migrations.AlterField(
            model_name='topic',
            name='subscribers',
            field=models.ManyToManyField(blank=True, related_name='topic_subscriptions', to=settings.AUTH_USER_MODEL, verbose_name='Subscribers'),
        ),
    ]



"""
    Forum tracking abstract models
    ==============================

    This module defines abstract models provided by the ``forum_tracking`` application.

"""

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from machina.core.loading import get_class


ForumReadTrackManager = get_class('forum_tracking.managers', 'ForumReadTrackManager')


class AbstractForumReadTrack(models.Model):
    """ Represents a track which records which forums have been read by a given user. """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name='forum_tracks', on_delete=models.CASCADE,
        verbose_name=_('User'),
    )
    forum = models.ForeignKey(
        'forum.Forum', related_name='tracks', on_delete=models.CASCADE, verbose_name=_('Forum'),
    )
    mark_time = models.DateTimeField(auto_now=True, db_index=True)

    objects = ForumReadTrackManager()

    class Meta:
        abstract = True
        app_label = 'forum_tracking'
        unique_together = ['user', 'forum', ]
        verbose_name = _('Forum track')
        verbose_name_plural = _('Forum tracks')

    def __str__(self):
        return '{} - {}'.format(self.user, self.forum)


class AbstractTopicReadTrack(models.Model):
    """ Represents a track which records which topics have been read by a given user. """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name='topic_tracks', on_delete=models.CASCADE,
        verbose_name=_('User'),
    )
    topic = models.ForeignKey(
        'forum_conversation.Topic', related_name='tracks', on_delete=models.CASCADE,
        verbose_name=_('Topic'),
    )
    mark_time = models.DateTimeField(auto_now=True, db_index=True)

    class Meta:
        abstract = True
        app_label = 'forum_tracking'
        unique_together = ['user', 'topic', ]
        verbose_name = _('Topic track')
        verbose_name_plural = _('Topic tracks')

    def __str__(self):
        return '{} - {}'.format(self.user, self.topic)



"""
    Django-machina settings
    =======================

    This module define the settings of the django-machina forum framework. Each setting can be
    overriden in the Django project's settings. These settings allow to customize many aspects of
    the forum application, such as conversations, polls, permissions, members, ...

"""

from django.conf import settings


# General
FORUM_NAME = getattr(settings, 'MACHINA_FORUM_NAME', 'Machina')
MARKUP_LANGUAGE = getattr(
    settings, 'MACHINA_MARKUP_LANGUAGE',
    ('machina.core.markdown.markdown', {'safe_mode': True, 'extras': {'break-on-newline': True}})
)
MARKUP_WIDGET = getattr(
    settings, 'MACHINA_MARKUP_WIDGET', 'machina.forms.widgets.MarkdownTextareaWidget'
)
MARKUP_WIDGET_KWARGS = getattr(settings, 'MACHINA_MARKUP_WIDGET_KWARGS', {})

MARKUP_MAX_LENGTH_VALIDATOR = getattr(
    settings, 'MACHINA_MARKUP_MAX_LENGTH_VALIDATOR',
    'machina.core.validators.NullableMaxLengthValidator'
)

BASE_TEMPLATE_NAME = getattr(settings, 'MACHINA_BASE_TEMPLATE_NAME', '_base.html')
USER_DISPLAY_NAME_METHOD = getattr(
    settings,
    'MACHINA_USER_DISPLAY_NAME_METHOD',
    'get_username',
)

# Forum
FORUM_IMAGE_UPLOAD_TO = getattr(settings, 'MACHINA_FORUM_IMAGE_UPLOAD_TO', 'machina/forum_images')
FORUM_IMAGE_WIDTH = getattr(settings, 'MACHINA_FORUM_IMAGE_WIDTH', 100)
FORUM_IMAGE_HEIGHT = getattr(settings, 'MACHINA_FORUM_IMAGE_HEIGHT', 70)

DEFAULT_FORUM_IMAGE_SETTINGS = {
    'width': FORUM_IMAGE_WIDTH,
    'height': FORUM_IMAGE_HEIGHT
}

FORUM_TOPICS_NUMBER_PER_PAGE = getattr(settings, 'MACHINA_FORUM_TOPICS_NUMBER_PER_PAGE', 20)


# Conversation
TOPIC_ANSWER_SUBJECT_PREFIX = getattr(settings, 'MACHINA_TOPIC_ANSWER_SUBJECT_PREFIX', 'Re:')
POST_CONTENT_MAX_LENGTH = getattr(settings, 'MACHINA_POST_CONTENT_MAX_LENGTH', None)

TOPIC_POSTS_NUMBER_PER_PAGE = getattr(settings, 'MACHINA_TOPIC_POSTS_NUMBER_PER_PAGE', 15)
TOPIC_REVIEW_POSTS_NUMBER = getattr(settings, 'MACHINA_TOPIC_REVIEW_POSTS_NUMBER', 10)


# Polls
POLL_MAX_OPTIONS_PER_POLL = getattr(settings, 'MACHINA_POLL_MAX_OPTIONS_PER_POLL', 30)
POLL_MAX_OPTIONS_PER_USER = getattr(settings, 'MACHINA_POLL_MAX_OPTIONS_PER_USER', 10)


# Attachments
ATTACHMENT_FILE_UPLOAD_TO = getattr(
    settings, 'MACHINA_ATTACHMENT_FILE_UPLOAD_TO', 'machina/attachments'
)
ATTACHMENT_CACHE_NAME = getattr(settings, 'MACHINA_ATTACHMENT_CACHE_NAME', 'machina_attachments')
ATTACHMENT_MAX_FILES_PER_POST = getattr(settings, 'MACHINA_ATTACHMENT_MAX_FILES_PER_POST', 15)

# Member
PROFILE_AVATAR_UPLOAD_TO = getattr(
    settings, 'MACHINA_PROFILE_AVATAR_UPLOAD_TO', 'machina/avatar_images'
)

PROFILE_AVATARS_ENABLED = getattr(settings, 'MACHINA_PROFILE_AVATARS_ENABLED', True)
PROFILE_AVATAR_WIDTH = getattr(settings, 'MACHINA_PROFILE_AVATAR_WIDTH', 150)
PROFILE_AVATAR_HEIGHT = getattr(settings, 'MACHINA_PROFILE_AVATAR_HEIGHT', 250)
PROFILE_AVATAR_MIN_WIDTH = getattr(settings, 'MACHINA_PROFILE_AVATAR_MIN_WIDTH', None)
PROFILE_AVATAR_MAX_WIDTH = getattr(settings, 'MACHINA_PROFILE_AVATAR_MAX_WIDTH', None)
PROFILE_AVATAR_MIN_HEIGHT = getattr(settings, 'MACHINA_PROFILE_AVATAR_MIN_HEIGHT', None)
PROFILE_AVATAR_MAX_HEIGHT = getattr(settings, 'MACHINA_PROFILE_AVATAR_MAX_HEIGHT', None)
PROFILE_AVATAR_MAX_UPLOAD_SIZE = getattr(settings, 'MACHINA_PROFILE_AVATAR_MAX_UPLOAD_SIZE', 0)

DEFAULT_AVATAR_SETTINGS = {
    'width': PROFILE_AVATAR_WIDTH,
    'height': PROFILE_AVATAR_HEIGHT,
    'min_width': PROFILE_AVATAR_MIN_WIDTH,
    'max_width': PROFILE_AVATAR_MAX_WIDTH,
    'min_height': PROFILE_AVATAR_MIN_HEIGHT,
    'max_height': PROFILE_AVATAR_MAX_HEIGHT,
    'max_upload_size': PROFILE_AVATAR_MAX_UPLOAD_SIZE
}

PROFILE_SIGNATURE_MAX_LENGTH = getattr(settings, 'MACHINA_PROFILE_SIGNATURE_MAX_LENGTH', 255)

PROFILE_RECENT_POSTS_NUMBER = getattr(settings, 'MACHINA_PROFILE_RECENT_POSTS_NUMBER', 15)
PROFILE_POSTS_NUMBER_PER_PAGE = getattr(settings, 'MACHINA_PROFILE_POSTS_NUMBER_PER_PAGE', 15)


# Permission
DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS = getattr(
    settings, 'MACHINA_DEFAULT_AUTHENTICATED_USER_FORUM_PERMISSIONS', []
)



"""
    Forum views
    ===========

    This module defines views provided by the ``forum`` application.

"""

from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.views.generic import ListView

from machina.apps.forum.signals import forum_viewed
from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')
Topic = get_model('forum_conversation', 'Topic')

ForumVisibilityContentTree = get_class('forum.visibility', 'ForumVisibilityContentTree')
PermissionRequiredMixin = get_class('forum_permission.viewmixins', 'PermissionRequiredMixin')
TrackingHandler = get_class('forum_tracking.handler', 'TrackingHandler')


class IndexView(ListView):
    """ Displays the top-level forums. """

    context_object_name = 'forums'
    template_name = 'forum/index.html'

    def get_queryset(self):
        """ Returns the list of items for this view. """
        return ForumVisibilityContentTree.from_forums(
            self.request.forum_permission_handler.forum_list_filter(
                Forum.objects.all(), self.request.user,
            ),
        )

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super(IndexView, self).get_context_data(**kwargs)
        visiblity_content_tree = context['forums']

        # Computes some global values.
        context['total_posts_count'] = sum(n.posts_count for n in visiblity_content_tree.top_nodes)
        context['total_topics_count'] = sum(
            n.topics_count for n in visiblity_content_tree.top_nodes
        )

        return context


class ForumView(PermissionRequiredMixin, ListView):
    """ Displays a forum and its topics. If applicable, its sub-forums can also be displayed. """

    context_object_name = 'topics'
    paginate_by = machina_settings.FORUM_TOPICS_NUMBER_PER_PAGE
    permission_required = ['can_read_forum', ]
    template_name = 'forum/forum_detail.html'
    view_signal = forum_viewed

    def get(self, request, **kwargs):
        """ Handles GET requests. """
        forum = self.get_forum()
        if forum.is_link:
            response = HttpResponseRedirect(forum.link)
        else:
            response = super(ForumView, self).get(request, **kwargs)
        self.send_signal(request, response, forum)
        return response

    def get_forum(self):
        """ Returns the forum to consider. """
        if not hasattr(self, 'forum'):
            self.forum = get_object_or_404(Forum, pk=self.kwargs['pk'])
        return self.forum

    def get_queryset(self):
        """ Returns the list of items for this view. """
        self.forum = self.get_forum()
        qs = (
            self.forum.topics
            .exclude(type=Topic.TOPIC_ANNOUNCE)
            .exclude(approved=False)
            .select_related('poster', 'last_post', 'last_post__poster')
        )
        return qs

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_forum()

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super(ForumView, self).get_context_data(**kwargs)

        # Insert the considered forum into the context
        context['forum'] = self.get_forum()

        # Get the list of forums that have the current forum as parent
        context['sub_forums'] = ForumVisibilityContentTree.from_forums(
            self.request.forum_permission_handler.forum_list_filter(
                context['forum'].get_descendants(), self.request.user,
            ),
        )

        # The announces will be displayed on each page of the forum
        context['announces'] = list(
            self.get_forum()
            .topics.select_related('poster', 'last_post', 'last_post__poster')
            .filter(type=Topic.TOPIC_ANNOUNCE)
        )

        # Determines the topics that have not been read by the current user
        context['unread_topics'] = TrackingHandler(self.request).get_unread_topics(
            list(context[self.context_object_name]) + context['announces'], self.request.user,
        )

        return context

    def send_signal(self, request, response, forum):
        """ Sends the signal associated with the view. """
        self.view_signal.send(
            sender=self, forum=forum, user=request.user, request=request, response=response,
        )



"""
    Forum permission middlewares
    ============================

    This module defines permission-related middlewares that can ease the process of verifying forum
    permissions during the request-response cycle.

"""

import uuid

from django.utils.deprecation import MiddlewareMixin

from machina.core.loading import get_class


PermissionHandler = get_class('forum_permission.handler', 'PermissionHandler')


class ForumPermissionMiddleware(MiddlewareMixin):
    """ This middleware attaches an instance of the PermissionHandler to each request.

    This allows to cache the permissions for the lifetime of the request object. The middleware also
    attaches a random identifier to each anonymous user in order to perform proper permission checks
    for anonymous users. This identifier is stored in the session.

    """

    anonymous_forum_key_session_id = '_anonymous_forum_key'

    def process_request(self, request):
        if not request.user.is_authenticated:
            # Get the anonymous forum key and attaches it the AnonymousUser instance.
            anonymous_forum_key = request.session.get(self.anonymous_forum_key_session_id, None)
            if anonymous_forum_key is None:
                anonymous_forum_key = self.get_anonymous_forum_key()
                request.session[self.anonymous_forum_key_session_id] = anonymous_forum_key

            setattr(request.user, 'forum_key', anonymous_forum_key)

        request.forum_permission_handler = PermissionHandler()

    def get_anonymous_forum_key(self):
        """ Returns a random anonymous forum key. """
        return uuid.uuid4().hex



"""
    Forum attachments forms
    =======================

    This module defines forms provided by the ``forum_attachments`` application.

"""

from django import forms
from django.forms.models import BaseModelFormSet, modelformset_factory

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model


Attachment = get_model('forum_attachments', 'Attachment')


class AttachmentForm(forms.ModelForm):
    """ Allows to upload forum attachments. """

    class Meta:
        model = Attachment
        fields = ['file', 'comment', ]


class BaseAttachmentFormset(BaseModelFormSet):
    """ Base class allowing to define forum attachment formsets. """

    def __init__(self, *args, **kwargs):
        self.post = kwargs.pop('post', None)
        super().__init__(*args, **kwargs)

    def save(self, commit=True, **kwargs):
        """ Saves the considered instances. """
        if self.post:
            for form in self.forms:
                form.instance.post = self.post
        super().save(commit)


AttachmentFormset = modelformset_factory(
    Attachment, AttachmentForm,
    formset=BaseAttachmentFormset,
    can_delete=True, extra=1,
    max_num=machina_settings.ATTACHMENT_MAX_FILES_PER_POST,
    validate_max=True,
)



from django import template

from machina.conf import settings as machina_settings


register = template.Library()


@register.filter
def posted_by(post, user):
    """ This will return a boolean indicating if the passed user has posted the given forum post.

    Usage::

        {% if post|posted_by:user %}...{% endif %}

    """
    return post.poster == user


@register.inclusion_tag('forum_conversation/topic_pages_inline_list.html')
def topic_pages_inline_list(topic):
    """ This will render an inline pagination for the posts related to the given topic.

    Usage::

        {% topic_pages_inline_list my_topic %}

    """
    data_dict = {
        'topic': topic,
    }

    pages_number = ((topic.posts_count - 1) // machina_settings.TOPIC_POSTS_NUMBER_PER_PAGE) + 1
    if pages_number > 5:
        data_dict['first_pages'] = range(1, 5)
        data_dict['last_page'] = pages_number
    elif pages_number > 1:
        data_dict['first_pages'] = range(1, pages_number + 1)

    return data_dict



# Generated by Django 2.1.2 on 2018-11-03 19:06

from django.db import migrations
import machina.apps.forum_member.abstract_models
import machina.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('forum_member', '0003_auto_20160227_2122'),
    ]

    operations = [
        migrations.AlterField(
            model_name='forumprofile',
            name='avatar',
            field=machina.models.fields.ExtendedImageField(blank=True, null=True, upload_to=machina.apps.forum_member.abstract_models.get_profile_avatar_upload_to, verbose_name='Avatar'),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _


class DatedModel(models.Model):
    """ Represents an abstract dated model.

    An abstract base class model that provides a created and a updated fields to store creation date
    and last updated date.

    """

    created = models.DateTimeField(auto_now_add=True, verbose_name=_('Creation date'))
    updated = models.DateTimeField(auto_now=True, verbose_name=_('Update date'))

    class Meta:
        abstract = True



"""
    Forum abstract models
    =====================

    This module defines abstract models provided by the ``forum`` application.

"""

import os
import uuid

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Sum
from django.utils.encoding import force_str
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from mptt.models import MPTTModel, TreeForeignKey

from machina.apps.forum import signals
from machina.conf import settings as machina_settings
from machina.models import DatedModel
from machina.models.fields import ExtendedImageField, MarkupTextField


def get_forum_image_upload_to(instance, filename):
    """ Returns a valid upload path for an image file associated with a forum instance. """
    return instance.get_image_upload_to(filename)


class AbstractForum(MPTTModel, DatedModel):
    """ The main forum model.

    The tree hierarchy of forums and categories is managed by the MPTTModel which is part of
    django-mptt.

    """

    parent = TreeForeignKey(
        'self', null=True, blank=True, related_name='children', on_delete=models.CASCADE,
        verbose_name=_('Parent'),
    )

    name = models.CharField(max_length=100, verbose_name=_('Name'))
    slug = models.SlugField(max_length=255, verbose_name=_('Slug'))

    description = MarkupTextField(verbose_name=_('Description'), null=True, blank=True)

    # A forum can come with an image (eg. a small logo)
    image = ExtendedImageField(
        null=True, blank=True, upload_to=get_forum_image_upload_to,
        verbose_name=_('Forum image'), **machina_settings.DEFAULT_FORUM_IMAGE_SETTINGS
    )

    # Forums can be simple links (eg. wiki, documentation, etc)
    link = models.URLField(verbose_name=_('Forum link'), null=True, blank=True)
    link_redirects = models.BooleanField(
        default=False, verbose_name=_('Track link redirects count'),
        help_text=_('Records the number of times a forum link was clicked'),
    )

    # Category, Default forum or Link ; that's what a forum can be
    FORUM_POST, FORUM_CAT, FORUM_LINK = 0, 1, 2
    TYPE_CHOICES = (
        (FORUM_POST, _('Default forum')),
        (FORUM_CAT, _('Category forum')),
        (FORUM_LINK, _('Link forum')),
    )
    type = models.PositiveSmallIntegerField(
        choices=TYPE_CHOICES, verbose_name=_('Forum type'), db_index=True,
    )

    # Tracking data (only approved topics and posts are recorded)
    direct_posts_count = models.PositiveIntegerField(
        editable=False, blank=True, default=0, verbose_name=_('Direct number of posts'),
    )
    direct_topics_count = models.PositiveIntegerField(
        editable=False, blank=True, default=0, verbose_name=_('Direct number of topics'),)
    link_redirects_count = models.PositiveIntegerField(
        editable=False, blank=True, default=0, verbose_name=_('Track link redirects count'),
    )

    # The 'last_post' and 'last_post_on' fields contain values related to the direct topics/posts
    # only (that is the topics/posts that are directly associated with the considered forum and not
    # one of its sub-forums).
    last_post = models.ForeignKey(
        'forum_conversation.Post', editable=False, related_name='+', blank=True, null=True,
        on_delete=models.SET_NULL, verbose_name=_('Last post'),
    )
    last_post_on = models.DateTimeField(verbose_name=_('Last post added on'), blank=True, null=True)

    # Display options ; these fields can be used to alter the display of the forums in the list of
    # forums.
    display_sub_forum_list = models.BooleanField(
        default=True, verbose_name=_('Display in parent-forums legend'),
        help_text=_('Displays this forum on the legend of its parent-forum (sub forums list)'),
    )

    class Meta:
        abstract = True
        app_label = 'forum'
        ordering = ['tree_id', 'lft']
        verbose_name = _('Forum')
        verbose_name_plural = _('Forums')

    def __str__(self):
        return self.name

    @property
    def margin_level(self):
        """ Returns a margin value computed from the forum node's level.

        Used in templates or menus to create an easy-to-see left margin to contrast a forum from
        their parents.

        """
        return self.level * 2

    @property
    def is_category(self):
        """ Returns ``True`` if the forum is a category. """
        return self.type == self.FORUM_CAT

    @property
    def is_forum(self):
        """ Returns ``True`` if the forum is a a default forum. """
        return self.type == self.FORUM_POST

    @property
    def is_link(self):
        """ Returns ``True`` if the forum is a link. """
        return self.type == self.FORUM_LINK

    def clean(self):
        """ Validates the forum instance. """
        super().clean()

        if self.parent and self.parent.is_link:
            raise ValidationError(_('A forum can not have a link forum as parent'))

        if self.is_category and self.parent and self.parent.is_category:
            raise ValidationError(_('A category can not have another category as parent'))

        if self.is_link and not self.link:
            raise ValidationError(_('A link forum must have a link associated with it'))

    def get_image_upload_to(self, filename):
        """ Returns the path to upload a new associated image to. """
        dummy, ext = os.path.splitext(filename)
        return os.path.join(
            machina_settings.FORUM_IMAGE_UPLOAD_TO,
            '{id}{ext}'.format(id=str(uuid.uuid4()).replace('-', ''), ext=ext),
        )

    def save(self, *args, **kwargs):
        """ Saves the forum instance. """
        # It is vital to track the changes of the parent associated with a forum in order to
        # maintain counters up-to-date and to trigger other operations such as permissions updates.
        old_instance = None
        if self.pk:
            old_instance = self.__class__._default_manager.get(pk=self.pk)

        # Update the slug field
        self.slug = slugify(force_str(self.name), allow_unicode=True)

        # Do the save
        super().save(*args, **kwargs)

        # If any change has been made to the forum parent, trigger the update of the counters
        if old_instance and old_instance.parent != self.parent:
            self.update_trackers()
            # Trigger the 'forum_moved' signal
            signals.forum_moved.send(sender=self, previous_parent=old_instance.parent)

    def update_trackers(self):
        """ Updates the denormalized trackers associated with the forum instance. """
        direct_approved_topics = self.topics.filter(approved=True).order_by('-last_post_on')

        # Compute the direct topics count and the direct posts count.
        self.direct_topics_count = direct_approved_topics.count()
        self.direct_posts_count = direct_approved_topics.aggregate(
            total_posts_count=Sum('posts_count'))['total_posts_count'] or 0

        # Forces the forum's 'last_post' ID and 'last_post_on' date to the corresponding values
        # associated with the topic with the latest post.
        if direct_approved_topics.exists():
            self.last_post_id = direct_approved_topics[0].last_post_id
            self.last_post_on = direct_approved_topics[0].last_post_on
        else:
            self.last_post_id = None
            self.last_post_on = None

        # Any save of a forum triggered from the update_tracker process will not result in checking
        # for a change of the forum's parent.
        self._simple_save()

    def _simple_save(self, *args, **kwargs):
        """ Simple wrapper around the standard save method.

        Calls the parent save method in order to avoid the checks for forum parent changes which can
        result in triggering a new update of the counters associated with the current forum.

        This allow the database to not be hit by such checks during very common and regular
        operations such as those provided by the update_trackers function; indeed these operations
        will never result in an update of a forum parent.

        """
        super().save(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('file', models.FileField(upload_to='machina/attachments', verbose_name='File')),
                ('comment', models.CharField(max_length=255, null=True, verbose_name='Comment', blank=True)),
                ('post', models.ForeignKey(related_name='attachments', verbose_name='Post', to='forum_conversation.Post', on_delete=models.CASCADE)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Attachment',
                'verbose_name_plural': 'Attachments',
            },
        ),
    ]



"""
    Forum member forms
    ==================

    This module defines forms provided by the ``forum_member`` application.

"""

from django import forms

from machina.conf import settings as machina_settings
from machina.core.db.models import get_model


ForumProfile = get_model('forum_member', 'ForumProfile')


class ForumProfileForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not machina_settings.PROFILE_AVATARS_ENABLED:
            del self.fields["avatar"]

    class Meta:
        model = ForumProfile
        fields = ['avatar', 'signature', ]



"""
    Forum permission abstract models
    ================================

    This module defines abstract models provided by the ``forum_permission`` application.

"""

from django.conf import settings
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from machina.core.loading import get_class


PermissionConfig = get_class('forum_permission.defaults', 'PermissionConfig')


class AbstractForumPermission(models.Model):
    """ Represents a single forum permission.

    The models that subclass ``AbstractForumPermission`` can be used to define forum permissions. A
    forum permission is basically defined by a codename and some booleans indicating if the
    considered permission can be granted globally (in that case the permission applies to all
    forums) or not.

    """

    codename = models.CharField(
        max_length=150, verbose_name=_('Permission codename'), unique=True, db_index=True,
    )

    class Meta:
        abstract = True
        app_label = 'forum_permission'
        verbose_name = _('Forum permission')
        verbose_name_plural = _('Forum permissions')

    def __str__(self):
        return '{} - {}'.format(self.codename, self.name)

    @cached_property
    def name(self):
        """ Returns the name of the considered permission. """
        perm_config = PermissionConfig().get(self.codename)
        return perm_config['label'] if perm_config else None


class BaseAuthForumPermission(models.Model):
    """ Represents a per-auth-component forum object permission. """

    permission = models.ForeignKey(
        'forum_permission.ForumPermission', on_delete=models.CASCADE,
        verbose_name=_('Forum permission'),
    )
    has_perm = models.BooleanField(verbose_name=_('Has perm'), default=True, db_index=True)

    # The forum related to a UserForumPermission instance can be null if the considered permission
    # should be granted globally.
    forum = models.ForeignKey(
        'forum.Forum', blank=True, null=True, on_delete=models.CASCADE, verbose_name=_('Forum'),
    )

    class Meta:
        abstract = True


class AbstractUserForumPermission(BaseAuthForumPermission):
    """ Represents a per-user forum object permission. """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.CASCADE,
        verbose_name=_('User'),
    )
    anonymous_user = models.BooleanField(
        verbose_name=_('Target anonymous user'), default=False, db_index=True,
    )
    authenticated_user = models.BooleanField(
        verbose_name=_('Target any logged in user'), default=False, db_index=True,
    )

    class Meta:
        abstract = True
        unique_together = ('permission', 'forum', 'user', )
        app_label = 'forum_permission'
        verbose_name = _('User forum permission')
        verbose_name_plural = _('User forum permissions')

    def __str__(self):
        if self.forum:
            return '{} - {} - {}'.format(self.permission, self.user, self.forum)
        return '{} - {}'.format(self.permission, self.user)

    def clean(self):
        """ Validates the current instance. """
        super().clean()
        if (
            (self.user is None and not self.anonymous_user and not self.authenticated_user) or
            ((self.user and self.anonymous_user) or (self.user and self.authenticated_user) or
             (self.anonymous_user and self.authenticated_user))
        ):
            raise ValidationError(
                _(
                    "A permission should target either a specific user, an anonymous user " +
                    "or any authenticated user"
                ),
            )


class AbstractGroupForumPermission(BaseAuthForumPermission):
    """ Represents a per-group forum object permission. """

    group = models.ForeignKey(Group, on_delete=models.CASCADE, verbose_name=_('Group'))

    class Meta:
        abstract = True
        unique_together = ('permission', 'forum', 'group', )
        app_label = 'forum_permission'
        verbose_name = _('Group forum permission')
        verbose_name_plural = _('Group forum permissions')

    def __str__(self):
        if self.forum:
            return '{} - {} - {}'.format(self.permission, self.group, self.forum)
        return '{} - {}'.format(self.permission, self.group)



from machina.conf import settings as machina_settings


def metadata(request):
    """ Appends some Machina-specific data to the template context. """
    return {
        'MACHINA_FORUM_NAME': machina_settings.FORUM_NAME,
        'MACHINA_BASE_TEMPLATE_NAME': machina_settings.BASE_TEMPLATE_NAME,
    }



"""
    Forum polls models
    ==================

    This module defines models provided by the ``forum_polls`` application.

"""

from machina.apps.forum_conversation.forum_polls.abstract_models import (
    AbstractTopicPoll, AbstractTopicPollOption, AbstractTopicPollVote
)
from machina.core.db.models import model_factory


TopicPoll = model_factory(AbstractTopicPoll)
TopicPollOption = model_factory(AbstractTopicPollOption)
TopicPollVote = model_factory(AbstractTopicPollVote)



"""
    Forum moderation URLs
    =====================

    This module defines URL patterns associated with the django-machina's ``forum_moderation``
    application.

"""

from django.urls import path

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumModerationURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_moderation`` application. """

    app_namespace = 'forum_moderation'

    topic_lock_view = get_class('forum_moderation.views', 'TopicLockView')
    topic_unlock_view = get_class('forum_moderation.views', 'TopicUnlockView')
    topic_delete_view = get_class('forum_moderation.views', 'TopicDeleteView')
    topic_move_view = get_class('forum_moderation.views', 'TopicMoveView')
    topic_update_to_normal_topic_view = get_class(
        'forum_moderation.views', 'TopicUpdateToNormalTopicView',
    )
    topic_update_to_sticky_topic_view = get_class(
        'forum_moderation.views', 'TopicUpdateToStickyTopicView',
    )
    topic_update_to_announce_view = get_class('forum_moderation.views', 'TopicUpdateToAnnounceView')
    moderation_queue_list_view = get_class('forum_moderation.views', 'ModerationQueueListView')
    moderation_queue_detail_view = get_class('forum_moderation.views', 'ModerationQueueDetailView')
    post_approve_view = get_class('forum_moderation.views', 'PostApproveView')
    post_disapprove_view = get_class('forum_moderation.views', 'PostDisapproveView')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path(
                'topic/<str:slug>-<int:pk>/lock/',
                self.topic_lock_view.as_view(),
                name='topic_lock',
            ),
            path(
                'topic/<str:slug>-<int:pk>/unlock/',
                self.topic_unlock_view.as_view(),
                name='topic_unlock',
            ),
            path(
                'topic/<str:slug>-<int:pk>/delete/',
                self.topic_delete_view.as_view(),
                name='topic_delete',
            ),
            path(
                'topic/<str:slug>-<int:pk>/move/',
                self.topic_move_view.as_view(),
                name='topic_move',
            ),
            path(
                'topic/<str:slug>-<int:pk>/change/topic/',
                self.topic_update_to_normal_topic_view.as_view(),
                name='topic_update_to_post',
            ),
            path(
                'topic/<str:slug>-<int:pk>/change/sticky/',
                self.topic_update_to_sticky_topic_view.as_view(),
                name='topic_update_to_sticky',
            ),
            path(
                'topic/<str:slug>-<int:pk>/change/announce/',
                self.topic_update_to_announce_view.as_view(),
                name='topic_update_to_announce',
            ),
            path('queue/', self.moderation_queue_list_view.as_view(), name='queue'),
            path(
                'queue/<int:pk>/',
                self.moderation_queue_detail_view.as_view(),
                name='queued_post',
            ),
            path(
                'queue/<int:pk>/approve/',
                self.post_approve_view.as_view(),
                name='approve_queued_post',
            ),
            path(
                'queue/<int:pk>/disapprove/',
                self.post_disapprove_view.as_view(),
                name='disapprove_queued_post',
            ),
        ]


urlpatterns_factory = ForumModerationURLPatternsFactory()



import pytest
from django.contrib.auth.models import User
from django.test.client import Client


@pytest.mark.django_db
class BaseClientTestCase:
    """
    Shortcut class for using Django's test client and avoid boilerplate code
    such as user login or user creation.
    """
    username = 'dummyuser'
    email = 'dummyuser@example.com'
    password = 'dummypassword'
    is_anonymous = False
    is_staff = False
    is_superuser = False
    permissions = []

    @pytest.fixture(autouse=True)
    def _setup_client(self):
        self.client = Client()
        if not self.is_anonymous:
            self.login()

    def login(self):
        self.user = self.create_user()
        self.client.login(username=self.username,
                          password=self.password)

    def create_user(self, username=None, password=None, email=None,
                    is_staff=None, is_superuser=None, permissions=None):
        user = User.objects.create_user(username or self.username,
                                        email or self.email,
                                        password or self.password)
        user.is_staff = is_staff if is_staff is not None else self.is_staff
        user.is_superuser = is_superuser if is_superuser is not None else self.is_superuser
        user.save()
        return user


class AdminClientTestCase(BaseClientTestCase):
    is_staff = True
    is_superuser = True



"""
    Forum permission handler
    ========================

    This module defines a ``PermissionHandler`` abstraction that allows to implement filter or
    access logic related to forums.

"""

import datetime as dt
from functools import reduce

from django.db import models
from django.utils.timezone import now
from mptt.utils import get_cached_trees

from machina.core.db.models import get_model
from machina.core.loading import get_class


Forum = get_model('forum', 'Forum')
GroupForumPermission = get_model('forum_permission', 'GroupForumPermission')
Post = get_model('forum_conversation', 'Post')
TopicPollVote = get_model('forum_polls', 'TopicPollVote')
UserForumPermission = get_model('forum_permission', 'UserForumPermission')

ForumPermissionChecker = get_class('forum_permission.checker', 'ForumPermissionChecker')

get_anonymous_user_forum_key = get_class(
    'forum_permission.shortcuts', 'get_anonymous_user_forum_key')


class PermissionHandler:
    """ Defines filter / access logic related to forums.

    The ``PermissionHandler`` class allows to filter lists of forums and to perform permission
    verifications on forums. It uses the ``ForumPermissionChecker`` class to perform these
    verifications.

    """

    def __init__(self):
        # This dictionary will store the forums that are granted for a specific lit of permission
        # codenames and a given user.
        self._granted_forums_cache = {}

        # This one will store the ancestors of the forums that will be used to perform permission
        # checks.
        self._forum_ancestors_cache = {}

        # This one will store ForumPermissionChecker instances for users whose permissions will be
        # checked.
        self._user_perm_checkers_cache = {}

    # Filtering methods
    # --

    def forum_list_filter(self, qs, user):
        """ Filters the given queryset in order to return a list of forums that can be seen and read
            by the specified user (at least).
        """
        # Any superuser should see all the forums
        if user.is_superuser:
            return qs

        # Check whether the forums can be viewed by the given user
        forums_to_hide = self._get_hidden_forum_ids(qs, user)

        return qs.exclude(id__in=forums_to_hide)

    def get_readable_forums(self, forums, user):
        """ Returns a queryset of forums that can be read by the considered user. """
        # Any superuser should be able to read all the forums.
        if user.is_superuser:
            return forums

        # Fetches the forums that can be read by the given user.
        readable_forums = self._get_forums_for_user(
            user, ['can_read_forum', ], use_tree_hierarchy=True)
        return forums.filter(id__in=[f.id for f in readable_forums]) \
            if isinstance(forums, (models.Manager, models.QuerySet)) \
            else list(filter(lambda f: f in readable_forums, forums))

    # Verification methods
    # --

    # Forums

    def can_read_forum(self, forum, user):
        """ Given a forum, checks whether the user can read its content. """
        return self._perform_basic_permission_check(forum, user, 'can_read_forum')

    # Posts and topics

    def can_add_topic(self, forum, user):
        """ Given a forum, checks whether the user can append topics to it. """
        return self._perform_basic_permission_check(forum, user, 'can_start_new_topics')

    def can_add_stickies(self, forum, user):
        """ Given a forum, checks whether the user can append stickies to it. """
        return self._perform_basic_permission_check(forum, user, 'can_post_stickies')

    def can_add_announcements(self, forum, user):
        """ Given a forum, checks whether the user can append announcements to it. """
        return self._perform_basic_permission_check(forum, user, 'can_post_announcements')

    def can_post_without_approval(self, forum, user):
        """ Given a forum, checks whether the user can add a posts and topics without approval. """
        return self._perform_basic_permission_check(forum, user, 'can_post_without_approval')

    def can_add_post(self, topic, user):
        """ Given a topic, checks whether the user can append posts to it. """
        can_add_post = self._perform_basic_permission_check(
            topic.forum, user, 'can_reply_to_topics',
        )
        can_add_post &= (
            not topic.is_locked or
            self._perform_basic_permission_check(topic.forum, user, 'can_reply_to_locked_topics')
        )
        return can_add_post

    def can_edit_post(self, post, user):
        """ Given a forum post, checks whether the user can edit the latter. """
        checker = self._get_checker(user)

        # A user can edit a post if...
        #     they are a superuser
        #     they are the original poster of the forum post
        #     they belong to the forum moderators
        is_author = self._is_post_author(post, user)
        can_edit = (
            user.is_superuser or
            (
                is_author and checker.has_perm('can_edit_own_posts', post.topic.forum) and
                not post.topic.is_locked
            ) or
            checker.has_perm('can_edit_posts', post.topic.forum)
        )
        return can_edit

    def can_delete_post(self, post, user):
        """
        Given a forum post, checks whether the user can delete the latter.
        """
        checker = self._get_checker(user)

        # A user can delete a post if...
        #     they are a superuser
        #     they are the original poster of the forum post
        #     they belong to the forum moderators
        is_author = self._is_post_author(post, user)
        can_delete = (
            user.is_superuser or
            (is_author and checker.has_perm('can_delete_own_posts', post.topic.forum)) or
            checker.has_perm('can_delete_posts', post.topic.forum)
        )
        return can_delete

    # Polls

    def can_create_polls(self, forum, user):
        """ Given a forum, checks whether the user can add a topic with an embedded poll. """
        return self._perform_basic_permission_check(forum, user, 'can_create_polls')

    def can_vote_in_poll(self, poll, user):
        """ Given a poll, checks whether the user can answer to it. """
        # First we have to check if the poll is curently open
        if poll.duration:
            poll_dtend = poll.created + dt.timedelta(days=poll.duration)
            if poll_dtend < now():
                return False

        # Is this user allowed to vote in polls in the current forum?
        can_vote = (
            self._perform_basic_permission_check(poll.topic.forum, user, 'can_vote_in_polls') and
            not poll.topic.is_locked
        )

        # Retrieve the user votes for the considered poll
        user_votes = TopicPollVote.objects.filter(poll_option__poll=poll)
        if user.is_anonymous:
            forum_key = get_anonymous_user_forum_key(user)
            if forum_key:
                user_votes = user_votes.filter(anonymous_key=forum_key)
            else:
                # If the forum key of the anonymous user cannot be retrieved, the user should not be
                # allowed to vote in the considered poll.
                user_votes = user_votes.none()
                can_vote = False
        else:
            user_votes = user_votes.filter(voter=user)

        # If the user has already voted, they can vote again if the vote changes are allowed
        if user_votes.exists() and can_vote:
            can_vote = poll.user_changes

        return can_vote

    # Attachments

    def can_attach_files(self, forum, user):
        """ Given a forum, checks whether the user can add attachments to posts. """
        return self._perform_basic_permission_check(forum, user, 'can_attach_file')

    def can_download_files(self, forum, user):
        """ Given a forum, checks whether the user can download files attached to posts. """
        return self._perform_basic_permission_check(forum, user, 'can_download_file')

    # Topic subscription

    def can_subscribe_to_topic(self, topic, user):
        """ Given a topic, checks whether the user can add it to their subscription list. """
        # A user can subscribe to topics if they are authenticated and if they have the permission
        # to read the related forum. Of course a user can subscribe only if they have not already
        # subscribed to the considered topic.
        return (
            user.is_authenticated and
            not topic.has_subscriber(user) and
            self._perform_basic_permission_check(topic.forum, user, 'can_read_forum')
        )

    def can_unsubscribe_from_topic(self, topic, user):
        """ Given a topic, checks whether the user can remove it from their subscription list. """
        # A user can unsubscribe from topics if they are authenticated and if they have the
        # permission to read the related forum. Of course a user can unsubscribe only if they are
        #  already a subscriber of the considered topic.
        return (
            user.is_authenticated and
            topic.has_subscriber(user) and
            self._perform_basic_permission_check(topic.forum, user, 'can_read_forum')
        )

    # Moderation

    def get_moderation_queue_forums(self, user):
        """ Returns the list of forums whose posts can be approved by the considered user. """
        return self._get_forums_for_user(user, ['can_approve_posts', ])

    def can_access_moderation_queue(self, user):
        """ Returns True if the passed user can access the moderation queue. """
        return len(self.get_moderation_queue_forums(user)) > 0

    def can_lock_topics(self, forum, user):
        """ Given a forum, checks whether the user can lock its topics. """
        return self._perform_basic_permission_check(forum, user, 'can_lock_topics')

    def can_move_topics(self, forum, user):
        """ Given a forum, checks whether the user can move its topics to another forum. """
        return self._perform_basic_permission_check(forum, user, 'can_move_topics')

    def get_target_forums_for_moved_topics(self, user):
        """ Returns a list of forums in which the considered user can add topics that have been
            moved from another forum.
        """
        return [f for f in self._get_forums_for_user(user, ['can_move_topics', ]) if f.is_forum]

    def can_delete_topics(self, forum, user):
        """ Given a forum, checks whether the user can delete its topics.

        Note: the ``can_delete_posts`` permission is used here because a user who can delete all the
        posts of a topic is also able to delete the topic itself.

        """
        return self._perform_basic_permission_check(forum, user, 'can_delete_posts')

    def can_update_topics_to_normal_topics(self, forum, user):
        """ Given a forum, checks whether the user can change its topic types to normal topics. """
        return self._perform_basic_permission_check(forum, user, 'can_edit_posts')

    def can_update_topics_to_sticky_topics(self, forum, user):
        """ Given a forum, checks whether the user can change its topic types to sticky topics. """
        return (
            self._perform_basic_permission_check(forum, user, 'can_edit_posts') and
            self._perform_basic_permission_check(forum, user, 'can_post_stickies')
        )

    def can_update_topics_to_announces(self, forum, user):
        """ Given a forum, checks whether the user can change its topic types to announces. """
        return (
            self._perform_basic_permission_check(forum, user, 'can_edit_posts') and
            self._perform_basic_permission_check(forum, user, 'can_post_announcements')
        )

    def can_approve_posts(self, forum, user):
        """ Given a forum, checks whether the user can approve its posts. """
        return self._perform_basic_permission_check(forum, user, 'can_approve_posts')

    # Common
    # --

    def _is_post_author(self, post, user):
        return (
            (post.poster == user) if user.is_authenticated else
            (
                post.anonymous_key is not None and
                post.anonymous_key == get_anonymous_user_forum_key(user)
            )
        )

    def _get_hidden_forum_ids(self, forums, user):
        """ Given a set of forums and a user, returns the list of forums that are not visible by
            this user.
        """
        visible_forums = self._get_forums_for_user(
            user, ['can_see_forum', 'can_read_forum', ], use_tree_hierarchy=True,
        )
        return forums.exclude(id__in=[f.id for f in visible_forums])

    def _get_forums_for_user(self, user, perm_codenames, use_tree_hierarchy=False):
        """ Returns all the forums that satisfy the given list of permission codenames.

        User and group forum permissions are used.

        If the ``use_tree_hierarchy`` keyword argument is set the granted forums will be filtered so
        that a forum which has an ancestor which is not in the granted forums set will not be
        returned.
        """
        granted_forums_cache_key = '{}__{}'.format(
            ':'.join(perm_codenames), user.id if not user.is_anonymous else 'anonymous',
        )

        if granted_forums_cache_key in self._granted_forums_cache:
            return self._granted_forums_cache[granted_forums_cache_key]

        forums = self._get_all_forums()

        # First check if the user is a superuser and if so, returns the forum queryset immediately.
        if user.is_superuser:  # pragma: no cover
            self._granted_forums_cache[granted_forums_cache_key] = forums
            return forums
        else:
            checker = self._get_checker(user)
            perms = checker.get_perms_for_forumlist(forums, perm_codenames)
            allowed_forums = []
            # Check if the requested permissions are in the set of permissions for the forum
            for f in forums:
                if set(perm_codenames).issubset(perms[f]):
                    allowed_forums.append(f)

            if use_tree_hierarchy:
                allowed_forums = self._filter_granted_forums_using_tree(allowed_forums)

        self._granted_forums_cache[granted_forums_cache_key] = allowed_forums
        return allowed_forums

    def _filter_granted_forums_using_tree(self, granted_forums):
        top_nodes = self._get_top_nodes()
        d = reduce(
            lambda a, b: a + self._filter_granted_node_using_tree(b, granted_forums), top_nodes, []
        )
        return Forum.objects.filter(id__in=d)

    def _filter_granted_node_using_tree(self, f, granted_forums):
        if f in granted_forums:
            return [f.id, ] + reduce(
                lambda a, b: a + self._filter_granted_node_using_tree(b, granted_forums),
                f.get_children(), []
            )
        return []

    def _get_top_nodes(self):
        if not hasattr(self, '_top_nodes'):
            self._top_nodes = get_cached_trees(Forum.objects.all())
        return self._top_nodes

    def _perform_basic_permission_check(self, forum, user, permission):
        """ Given a forum and a user, checks whether the latter has the passed permission.

        The workflow is:

            1. The permission is granted if the user is a superuser
            2. If not, a check is performed with the given permission

        """
        checker = self._get_checker(user)

        # The action is granted if...
        #     the user is the superuser
        #     the user has the permission to do so
        check = (user.is_superuser or checker.has_perm(permission, forum))
        return check

    def _get_checker(self, user):
        """ Return a ForumPermissionChecker instance for the given user. """
        user_perm_checkers_cache_key = user.id if not user.is_anonymous else 'anonymous'

        if user_perm_checkers_cache_key in self._user_perm_checkers_cache:
            return self._user_perm_checkers_cache[user_perm_checkers_cache_key]

        checker = ForumPermissionChecker(user)
        self._user_perm_checkers_cache[user_perm_checkers_cache_key] = checker
        return checker

    def _get_all_forums(self):
        """ Returns all forums. """
        if not hasattr(self, '_all_forums'):
            self._all_forums = list(Forum.objects.all())
        return self._all_forums



"""
    Forum tracking managers
    =======================

    This module defines managers provided by the ``forum_tracking`` application.

"""

from django.db import models

from machina.core.loading import get_class


ForumVisibilityContentTree = get_class('forum.visibility', 'ForumVisibilityContentTree')


class ForumReadTrackManager(models.Manager):
    """ Provides useful manager methods for the ``ForumReadTrack`` model. """

    def get_unread_forums_from_list(self, forums, user):
        """ Filter a list of forums and return only those which are unread.

        Given a list of forums find and returns the list of forums that are unread for the passed
        user. If a forum is unread all of its ancestors are also unread and will be included in the
        final list.
        """
        unread_forums = []
        visibility_contents = ForumVisibilityContentTree.from_forums(forums)
        forum_ids_to_visibility_nodes = visibility_contents.as_dict

        tracks = super().get_queryset().select_related('forum').filter(
            user=user,
            forum__in=forums)
        tracked_forums = []

        for track in tracks:
            forum_last_post_on = forum_ids_to_visibility_nodes[track.forum_id].last_post_on
            if (forum_last_post_on and track.mark_time < forum_last_post_on) \
                    and track.forum not in unread_forums:
                unread_forums.extend(track.forum.get_ancestors(include_self=True))
            tracked_forums.append(track.forum)

        for forum in forums:
            if forum not in tracked_forums and forum not in unread_forums \
                    and forum.direct_topics_count > 0:
                unread_forums.extend(forum.get_ancestors(include_self=True))

        return list(set(unread_forums))



"""
    Forum attachments cache
    =======================

    This module defines an abstraction allowing to put forum attachments into cache when users are
    creating forum posts and topics.

"""

from io import BytesIO

from django.conf import settings
from django.core.cache import InvalidCacheBackendError, caches
from django.core.exceptions import ImproperlyConfigured
from django.core.files.uploadedfile import InMemoryUploadedFile, TemporaryUploadedFile
from django.utils.datastructures import MultiValueDict

from machina.conf import settings as machina_settings


class AttachmentCache:
    """ The attachments cache.

    This one should be used with a FileBasedCache backend. But this can be overriden. The
    attachments cache acts as a wrapper and ensure that the states (name, size, content type,
    charset and content) of all files from any request.FILES dict are saved inside the considered
    backend when calling the 'set' method. Conversely, the 'get' method will populate a dictionary
    of InMemoryUploadedFile instances or TemporaryUploadedFile instancesby using these states.

    """
    def __init__(self):
        self.backend = self.get_backend()

    def get_backend(self):
        """ Returns the associated cache backend. """
        try:
            cache = caches[machina_settings.ATTACHMENT_CACHE_NAME]
        except InvalidCacheBackendError:
            raise ImproperlyConfigured(
                'The attachment cache backend ({}) is not configured'.format(
                    machina_settings.ATTACHMENT_CACHE_NAME,
                ),
            )
        return cache

    def set(self, key, files):
        """ Stores the state of each file embedded in the request.FILES MultiValueDict instance.

        This instance is assumed to be passed as the 'files' argument. Each state stored in the
        cache is a dictionary containing the following values:

            name
                The name of the uploaded file.
            size
                The size of the uploaded file.
            content_type
                The content type of the uploaded file.
            content_length
                The content length of the uploaded file.
            charset
                The charset of the uploaded file.
            content
                The content of the uploaded file.

        """
        files_states = {}

        for name, upload in files.items():
            # Generates the state of the file
            state = {
                'name': upload.name,
                'size': upload.size,
                'content_type': upload.content_type,
                'charset': upload.charset,
                'content': upload.file.read(),
            }
            files_states[name] = state

            # Go to the first byte in the file for future use
            upload.file.seek(0)
        self.backend.set(key, files_states)

    def get(self, key):
        """ Regenerates a MultiValueDict instance containing the files related to all file states
            stored for the given key.
        """
        upload = None
        files_states = self.backend.get(key)
        files = MultiValueDict()
        if files_states:
            for name, state in files_states.items():
                f = BytesIO()
                f.write(state['content'])

                # If the post is too large, we cannot use a
                # InMemoryUploadedFile instance.
                if state['size'] > settings.FILE_UPLOAD_MAX_MEMORY_SIZE:
                    upload = TemporaryUploadedFile(
                        state['name'],
                        state['content_type'],
                        state['size'],
                        state['charset'],
                    )
                    upload.file = f
                else:
                    f = BytesIO()
                    f.write(state['content'])
                    upload = InMemoryUploadedFile(
                        file=f,
                        field_name=name,
                        name=state['name'],
                        content_type=state['content_type'],
                        size=state['size'],
                        charset=state['charset'],
                    )
                files[name] = upload

                # Go to the first byte in the file for future use
                upload.file.seek(0)

        return files

    def delete(self, key):
        """ Deletes a file associated with a specific key. """
        self.backend.delete(key)


cache = AttachmentCache()



# Generated by Django 2.2.2 on 2019-06-28 02:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum', '0010_auto_20181103_1401'),
    ]

    operations = [
        migrations.AlterField(
            model_name='forum',
            name='level',
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name='forum',
            name='lft',
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name='forum',
            name='rght',
            field=models.PositiveIntegerField(editable=False),
        ),
    ]



"""
    Forum polls app config
    ======================

    This module contains the application configuration class - available in the Django app registry.
    For more information on this file, see https://docs.djangoproject.com/en/dev/ref/applications/

"""

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class ForumPollsAppConfig(AppConfig):
    label = 'forum_polls'
    name = 'machina.apps.forum_conversation.forum_polls'
    verbose_name = _('Machina: Forum polls')



"""
    Forum conversation models
    =========================

    This module defines models provided by the ``forum_conversation`` application.

"""

from machina.apps.forum_conversation.abstract_models import AbstractPost, AbstractTopic
from machina.core.db.models import model_factory


Topic = model_factory(AbstractTopic)
Post = model_factory(AbstractPost)



"""
    Forum member model admin definitions
    ====================================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from django.contrib import admin

from machina.core.db.models import get_model
from machina.models.fields import MarkupTextField, MarkupTextFieldWidget


ForumProfile = get_model('forum_member', 'ForumProfile')


class ForumProfileAdmin(admin.ModelAdmin):
    """ The Forum Profile model admin. """

    list_display = ('id', 'user', 'posts_count', )
    list_filter = ('posts_count', )
    list_display_links = ('id', 'user', )
    raw_id_fields = ('user', )
    search_fields = ('user__username',)

    formfield_overrides = {
        MarkupTextField: {'widget': MarkupTextFieldWidget},
    }


admin.site.register(ForumProfile, ForumProfileAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.2 on 2016-02-25 04:15
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations
import machina.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('forum_member', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='forumprofile',
            name='signature',
            field=machina.models.fields.MarkupTextField(blank=True, no_rendered_field=True, null=True, validators=[django.core.validators.MaxLengthValidator(255)], verbose_name='Signature'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-09-03 03:12
from __future__ import unicode_literals

from django.db import migrations


def update_topic_first_post_last_post(apps, schema_editor):
    Topic = apps.get_model('forum_conversation', 'Topic')
    for topic in Topic.objects.all():
        first_post = topic.posts.all().order_by('created').first()
        last_post = topic.posts.filter(approved=True).order_by('-created').first()
        topic.first_post = first_post
        topic.last_post = last_post
        topic.save()

def reverse_topic_first_post_last_post(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0007_auto_20160903_0450'),
    ]

    operations = [
        migrations.RunPython(
            update_topic_first_post_last_post, reverse_code=reverse_topic_first_post_last_post),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('forum_polls', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='topicpollvote',
            name='anonymous_key',
            field=models.CharField(max_length=100, null=True, verbose_name='Anonymous user forum key', blank=True),
        ),
        migrations.AlterField(
            model_name='topicpollvote',
            name='voter',
            field=models.ForeignKey(related_name='poll_votes', verbose_name='Voter', blank=True, to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE),
        ),
    ]



"""
    The visibility module
    =====================

    This module provides helper classes to manage and compute values that should be displayed when
    considering a tree of forums. This includes post counts, topic counts, siblings, ...

"""

from django.db.models.query import QuerySet
from django.utils.functional import cached_property


class ForumVisibilityContentTree:
    """ Represents a tree of ``ForumVisibilityContentNode`` instances.

    Such a tree can be used to easily compute sums or "global" values associated with a given set of
    forum instances. It can be useful to display forum information to a user (eg. in a list of
    forums, etc.).

    """

    def __init__(self, nodes=None):
        # The "nodes" argument should contain a list of ForumVisibilityContentNode instances. It
        # should be a list of all the nodes contained in the considered tree (regardless their
        # position in the tree of forum).
        self.nodes = nodes or []

    def __bool__(self):
        return len(self.forums) > 0

    @classmethod
    def from_forums(cls, forums):
        """ Initializes a ``ForumVisibilityContentTree`` instance from a list of forums. """
        root_level = None
        current_path = []
        nodes = []

        # Ensures forums last posts and related poster relations are "followed" for better
        # performance (only if we're considering a queryset).
        forums = (
            forums.select_related('last_post', 'last_post__poster')
            if isinstance(forums, QuerySet) else forums
        )

        for forum in forums:
            level = forum.level

            # Set the root level to the top node level at the first iteration.
            if root_level is None:
                root_level = level

            # Initializes a visibility forum node associated with current forum instance.
            vcontent_node = ForumVisibilityContentNode(forum)

            # Computes a relative level associated to the node.
            relative_level = level - root_level
            vcontent_node.relative_level = relative_level

            # All children nodes will be stored in an array attached to the current node.
            vcontent_node.children = []

            # Removes the forum that are not in the current branch.
            while len(current_path) > relative_level:
                current_path.pop(-1)

            if level != root_level:
                # Update the parent of the current forum.
                parent_node = current_path[-1]
                vcontent_node.parent = parent_node
                parent_node.children.append(vcontent_node)

            # Sets visible flag if applicable. The visible flag is used to determine whether a forum
            # can be seen in a forum list or not. A forum can be seen if one of the following
            # statements is true:
            #
            # * the forum is a direct child of the starting forum for the considered level
            # * the forum have a parent which is a category and this category is a direct child of
            #   the starting forum
            # * the forum have its 'display_sub_forum_list' option set to True and have a parent
            #   which is another forum. The latter is a direct child of the starting forum
            # * the forum have its 'display_sub_forum_list' option set to True and have a parent
            #   which is another forum. The later have a parent which is a category and this
            #   category is a direct child of the starting forum
            #
            # If forums at the root level don't have parents, the visible forums are those that can
            # be seen from the root of the forums tree.
            vcontent_node.visible = (
                (relative_level == 0) or
                (forum.display_sub_forum_list and relative_level == 1) or
                (forum.is_category and relative_level == 1) or
                (
                    relative_level == 2 and
                    vcontent_node.parent.parent.obj.is_category and
                    vcontent_node.parent.obj.is_forum
                )
            )

            # Add the current forum to the end of the current branch and inserts the node inside the
            # final node dictionary.
            current_path.append(vcontent_node)
            nodes.append(vcontent_node)

        tree = cls(nodes=nodes)
        for node in tree.nodes:
            node.tree = tree

        return tree

    @cached_property
    def as_dict(self):
        """ Returns a dictionary of forum ID / related node. """
        return {n.obj.id: n for n in self.nodes}

    @cached_property
    def forums(self):
        """ Returns a list of ``Forum`` instances associated with the underlying nodes. """
        return [n.obj for n in self.nodes]

    @cached_property
    def root_level(self):
        """ Returns the root level of the considered tree. """
        return self.top_nodes[0].level if self.top_nodes else None

    @cached_property
    def top_nodes(self):
        """ Returns only the node without immediate parent. """
        return list(filter(lambda n: not n.relative_level, self.nodes))

    @cached_property
    def visible_forums(self):
        """ Returns only the forum instances associated with the current tree. """
        return [n.obj for n in self.visible_nodes]

    @cached_property
    def visible_nodes(self):
        """ Returns only the visible nodes associated with the current tree. """
        return list(filter(lambda n: n.visible, self.nodes))


class ForumVisibilityContentNode:
    """ Represents a forum object and its "visibility content".

    This class provides common properties that should help computing values such as posts counts or
    topics counts for a specific forum instance.
    """

    def __init__(self, obj):
        self.obj = obj
        self.level = obj.level
        self.relative_level = None
        self.parent = None
        self.children = []
        self.tree = None
        self.visible = False

    @cached_property
    def last_post(self):
        """ Returns the latest post associated with the node or one of its descendants. """
        posts = [n.last_post for n in self.children if n.last_post is not None]
        children_last_post = max(posts, key=lambda p: p.created) if posts else None
        if children_last_post and self.obj.last_post_id:
            return max(self.obj.last_post, children_last_post, key=lambda p: p.created)
        return children_last_post or self.obj.last_post

    @cached_property
    def last_post_on(self):
        """ Returns the latest post date associated with the node or one of its descendants. """
        dates = [n.last_post_on for n in self.children if n.last_post_on is not None]
        children_last_post_on = max(dates) if dates else None
        if children_last_post_on and self.obj.last_post_on:
            return max(self.obj.last_post_on, children_last_post_on)
        return children_last_post_on or self.obj.last_post_on

    @cached_property
    def next_sibling(self):
        """ Returns the next sibling of the current node.

        The next sibling is searched in the parent node if we are not considering a top-level node.
        Otherwise it is searched inside the list of nodes (which should be sorted by tree ID) that
        is associated with the considered tree instance.
        """
        if self.parent:
            nodes = self.parent.children
            index = nodes.index(self)
            sibling = nodes[index + 1] if index < len(nodes) - 1 else None
        else:
            nodes = self.tree.nodes
            index = nodes.index(self)
            sibling = (
                next((n for n in nodes[index + 1:] if n.level == self.level), None)
                if index < len(nodes) - 1 else None
            )
        return sibling

    @cached_property
    def posts_count(self):
        """ Returns the number of posts associated with the current node and its descendants. """
        return self.obj.direct_posts_count + sum(n.posts_count for n in self.children)

    @cached_property
    def previous_sibling(self):
        """ Returns the previous sibling of the current node.

        The previous sibling is searched in the parent node if we are not considering a top-level
        node. Otherwise it is searched inside the list of nodes (which should be sorted by tree ID)
        that is associated with the considered tree instance.
        """
        if self.parent:
            nodes = self.parent.children
            index = nodes.index(self)
            sibling = nodes[index - 1] if index > 0 else None
        else:
            nodes = self.tree.nodes
            index = nodes.index(self)
            sibling = (
                next((n for n in reversed(nodes[:index]) if n.level == self.level), None)
                if index > 0 else None
            )
        return sibling

    @cached_property
    def topics_count(self):
        """ Returns the number of topics associated with the current node and its descendants. """
        return self.obj.direct_topics_count + sum(n.topics_count for n in self.children)



"""
    Forum models
    ============

    This module defines models provided by the ``forum`` application.

"""

from machina.apps.forum.abstract_models import AbstractForum
from machina.core.db.models import model_factory


Forum = model_factory(AbstractForum)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['machina',
 'machina.apps',
 'machina.apps.forum',
 'machina.apps.forum.migrations',
 'machina.apps.forum_conversation',
 'machina.apps.forum_conversation.forum_attachments',
 'machina.apps.forum_conversation.forum_attachments.migrations',
 'machina.apps.forum_conversation.forum_polls',
 'machina.apps.forum_conversation.forum_polls.migrations',
 'machina.apps.forum_conversation.migrations',
 'machina.apps.forum_feeds',
 'machina.apps.forum_member',
 'machina.apps.forum_member.migrations',
 'machina.apps.forum_moderation',
 'machina.apps.forum_permission',
 'machina.apps.forum_permission.migrations',
 'machina.apps.forum_search',
 'machina.apps.forum_tracking',
 'machina.apps.forum_tracking.migrations',
 'machina.conf',
 'machina.core',
 'machina.core.db',
 'machina.forms',
 'machina.models',
 'machina.templatetags',
 'machina.test',
 'machina.test.factories']

package_data = \
{'': ['*'],
 'machina': ['locale/ar/LC_MESSAGES/*',
             'locale/ca/LC_MESSAGES/*',
             'locale/de/LC_MESSAGES/*',
             'locale/en/LC_MESSAGES/*',
             'locale/es/LC_MESSAGES/*',
             'locale/fr/LC_MESSAGES/*',
             'locale/fr_CA/LC_MESSAGES/*',
             'locale/hu/LC_MESSAGES/*',
             'locale/it/LC_MESSAGES/*',
             'locale/ja_JP/LC_MESSAGES/*',
             'locale/nb/LC_MESSAGES/*',
             'locale/nl_NL/LC_MESSAGES/*',
             'locale/pl/LC_MESSAGES/*',
             'locale/pt_BR/LC_MESSAGES/*',
             'locale/ro/LC_MESSAGES/*',
             'locale/ro_RO/LC_MESSAGES/*',
             'locale/ru/LC_MESSAGES/*',
             'locale/sl_SI/LC_MESSAGES/*',
             'locale/zh/LC_MESSAGES/*',
             'locale/zh_CN/LC_MESSAGES/*',
             'locale/zh_TW/LC_MESSAGES/*',
             'static/machina/build/*',
             'static/machina/build/css/*',
             'static/machina/build/css/vendor/*',
             'static/machina/build/fonts/*',
             'static/machina/build/js/*',
             'static/machina/build/js/vendor/*',
             'static/machina/js/*',
             'static/machina/sass/*',
             'static/machina/sass/admin_pages/*',
             'static/machina/sass/board_pages/*',
             'templates/admin/forum/forum/*',
             'templates/machina/*',
             'templates/machina/forum/*',
             'templates/machina/forum_conversation/*',
             'templates/machina/forum_conversation/forum_attachments/*',
             'templates/machina/forum_conversation/forum_polls/*',
             'templates/machina/forum_conversation/partials/*',
             'templates/machina/forum_feeds/*',
             'templates/machina/forum_member/*',
             'templates/machina/forum_moderation/*',
             'templates/machina/forum_moderation/moderation_queue/*',
             'templates/machina/forum_search/*',
             'templates/machina/forum_tracking/*',
             'templates/machina/partials/*']}

install_requires = \
['django-haystack>=2.1',
 'django-mptt>=0.10.0',
 'django-widget-tweaks>=1.4',
 'django>=2.2',
 'markdown2>=2.3',
 'pillow>=2.2']

setup_kwargs = {
    'name': 'django-machina',
    'version': '1.1.2',
    'description': 'A Django forum engine for building powerful community driven websites.',
    'long_description': 'Django-machina\n##############\n\nDjango-machina is a forum framework for Django providing a way to build community-driven websites.\nIt offers a full-featured yet very extensible forum solution:\n\n* Topic and post editing\n* Forums tree management\n* Per-forum permissions\n* Anonymous posting\n* Polls and attachments\n* Moderation and pre-moderation\n* Forum conversations search\n* ...\n\nDjango-machina was built with integration in mind: the application is designed to be used inside\nexisting Django applications. It is not a standalone forum solution.\n\nDjango-machina was built with customization and extensibility in mind: each single functionality\nof the application can be customized or overriden to accommodate with your needs.\n\n.. contents:: Table of Contents\n    :local:\n\nDocumentation\n=============\n\nOnline browsable documentation is available at https://django-machina.readthedocs.org.\n\nHead over to the documentation for all the details on how to set up your forum and how to customize\nit to suit your needs.\n\nRequirements\n============\n\nPython 3.5+, Django 2.2+. Please refer to the requirements_ section of the documentation for a full\nlist of dependencies.\n\n.. _requirements: https://django-machina.readthedocs.org/en/stable/getting_started.html#requirements\n\nDemo site\n=========\n\nA demo site can be tested at https://demo.machina-forum.io. It contains a standard installation\nof django-machina without any customizations.\n\nAuthors\n=======\n\nMorgan Aubert (`@ellmetha <https://github.com/ellmetha>`_) and contributors_. See ``AUTHORS`` for\nmore details.\n\n.. _contributors: https://github.com/ellmetha/django-machina/contributors\n\nLicense\n=======\n\nBSD. See ``LICENSE`` for more details.\n',
    'author': 'Morgan Aubert',
    'author_email': 'me@morganaubert.name',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ellmetha/django-machina',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



"""
    Forum conversation model admin definitions
    ==========================================

    This module defines admin classes used to populate the Django administration dashboard.

"""

from django.contrib import admin

from machina.core.db.models import get_model
from machina.models.fields import MarkupTextField, MarkupTextFieldWidget


Attachment = get_model('forum_attachments', 'Attachment')
Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')


class AttachmentInline(admin.TabularInline):
    model = Attachment
    extra = 1


class PostAdmin(admin.ModelAdmin):
    """ The Post model admin. """

    inlines = [AttachmentInline, ]
    list_display = ('__str__', 'topic', 'poster', 'updated', 'approved')
    list_filter = ('created', 'updated',)
    raw_id_fields = ('poster', 'topic',)
    search_fields = ('content',)
    list_editable = ('approved',)

    formfield_overrides = {
        MarkupTextField: {'widget': MarkupTextFieldWidget},
    }


class TopicAdmin(admin.ModelAdmin):
    """ The Topic model admin. """

    list_display = (
        'subject', 'forum', 'created', 'first_post', 'last_post', 'posts_count', 'approved',
    )
    list_filter = ('created', 'updated',)
    raw_id_fields = ('poster', 'subscribers', )
    search_fields = ('subject',)
    list_editable = ('approved',)


admin.site.register(Topic, TopicAdmin)
admin.site.register(Post, PostAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-09 04:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forum_conversation', '0005_auto_20160607_0455'),
    ]

    operations = [
        migrations.AddField(
            model_name='post',
            name='enable_signature',
            field=models.BooleanField(db_index=True, default=True, verbose_name='Attach a signature'),
        ),
    ]



"""
    Forum conversation views
    ========================

    This module defines views provided by the ``forum_conversation`` application.

"""

from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DeleteView, FormView, ListView
from django.views.generic.detail import SingleObjectMixin

from machina.apps.forum_conversation.signals import topic_viewed
from machina.conf import settings as machina_settings
from machina.core.db.models import get_model
from machina.core.loading import get_class


Attachment = get_model('forum_attachments', 'Attachment')
Forum = get_model('forum', 'Forum')
Post = get_model('forum_conversation', 'Post')
Topic = get_model('forum_conversation', 'Topic')
TopicPollOption = get_model('forum_polls', 'TopicPollOption')

AttachmentFormset = get_class('forum_attachments.forms', 'AttachmentFormset')
PostForm = get_class('forum_conversation.forms', 'PostForm')
TopicForm = get_class('forum_conversation.forms', 'TopicForm')
TopicPollOptionFormset = get_class('forum_polls.forms', 'TopicPollOptionFormset')
TopicPollVoteForm = get_class('forum_polls.forms', 'TopicPollVoteForm')

attachments_cache = get_class('forum_attachments.cache', 'cache')

PermissionRequiredMixin = get_class('forum_permission.viewmixins', 'PermissionRequiredMixin')


class TopicView(PermissionRequiredMixin, ListView):
    """ Displays a forum topic. """

    context_object_name = 'posts'
    paginate_by = machina_settings.TOPIC_POSTS_NUMBER_PER_PAGE
    permission_required = ['can_read_forum', ]
    poll_form_class = TopicPollVoteForm
    template_name = 'forum_conversation/topic_detail.html'
    view_signal = topic_viewed

    def get(self, request, **kwargs):
        """ Handles GET requests. """
        topic = self.get_topic()

        # Handle pagination
        requested_post = request.GET.get('post', None)
        if requested_post:
            try:
                assert requested_post.isdigit()
                post = topic.posts.get(pk=requested_post)
                requested_page = (
                    ((post.position - 1) // machina_settings.TOPIC_POSTS_NUMBER_PER_PAGE) + 1
                )
                request.GET = request.GET.copy()  # A QueryDict is immutable
                request.GET.update({'page': requested_page})
            except (Post.DoesNotExist, AssertionError):
                pass

        response = super().get(request, **kwargs)
        self.send_signal(request, response, topic)
        return response

    def get_topic(self):
        """ Returns the topic to consider. """
        if not hasattr(self, 'topic'):
            self.topic = get_object_or_404(
                Topic.objects.select_related('forum').all(), pk=self.kwargs['pk'],
            )
        return self.topic

    def get_queryset(self):
        """ Returns the list of items for this view. """
        self.topic = self.get_topic()
        qs = (
            self.topic.posts
            .all()
            .exclude(approved=False)
            .select_related('poster', 'updated_by')
            .prefetch_related('attachments', 'poster__forum_profile')
        )
        return qs

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_topic().forum

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)

        # Insert the considered topic and the associated forum into the context
        topic = self.get_topic()
        context['topic'] = topic
        context['forum'] = topic.forum

        # Handles the case when a poll is associated to the topic
        try:
            if hasattr(topic, 'poll') and topic.poll.options.exists():
                context['poll'] = topic.poll
                context['poll_form'] = self.poll_form_class(poll=topic.poll)
                context['view_results_action'] = self.request.GET.get('view_results', None)
                context['change_vote_action'] = self.request.GET.get('change_vote', None)
        except ObjectDoesNotExist:  # pragma: no cover
            pass

        return context

    def send_signal(self, request, response, topic):
        """ Sends the signal associated with the view. """
        self.view_signal.send(
            sender=self, topic=topic, user=request.user, request=request, response=response,
        )


class BasePostFormView(FormView):
    """ A base view for handling post forms. """

    approval_required_message = _('This message will be validated before appearing on the forum.')
    attachment_formset_class = AttachmentFormset
    attachment_formset_general_error_message = _(
        'There are some errors in the attachments you submitted.'
    )
    forum_pk_url_kwarg = None
    post_form_class = PostForm
    post_pk_url_kwarg = None
    success_message = _('This message has been posted successfully.')
    topic_pk_url_kwarg = None

    def get(self, request, *args, **kwargs):
        """ Handles GET requests. """
        self.init_attachment_cache()

        # Initializes the forms
        post_form_class = self.get_post_form_class()
        post_form = self.get_post_form(post_form_class)
        attachment_formset_class = self.get_attachment_formset_class()
        attachment_formset = self.get_attachment_formset(attachment_formset_class)

        return self.render_to_response(
            self.get_context_data(post_form=post_form, attachment_formset=attachment_formset),
        )

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        self.init_attachment_cache()

        # Stores a boolean indicating if we are considering a preview
        self.preview = 'preview' in self.request.POST

        # Initializes the forms
        post_form_class = self.get_post_form_class()
        post_form = self.get_post_form(post_form_class)
        attachment_formset_class = self.get_attachment_formset_class()
        attachment_formset = self.get_attachment_formset(attachment_formset_class)

        self.attachment_preview = (
            self.preview if attachment_formset and attachment_formset.is_valid() else None
        )

        post_form_valid = post_form.is_valid()
        if (
            (post_form_valid and attachment_formset is None) or
            (post_form_valid and attachment_formset.is_valid())
        ):
            return self.form_valid(post_form, attachment_formset)
        else:
            return self.form_invalid(post_form, attachment_formset)

    def init_attachment_cache(self):
        """ Initializes the attachment cache for the current view. """
        if self.request.method == 'GET':
            # Invalidates previous attachments
            attachments_cache.delete(self.get_attachments_cache_key(self.request))
            return

        # Try to restore previous uploaded attachments if applicable
        attachments_cache_key = self.get_attachments_cache_key(self.request)
        restored_attachments_dict = attachments_cache.get(attachments_cache_key)
        if restored_attachments_dict:
            restored_attachments_dict.update(self.request.FILES)
            self.request._files = restored_attachments_dict

        # Updates the attachment cache if files are available
        if self.request.FILES:
            attachments_cache.set(attachments_cache_key, self.request.FILES)

    def get_attachments_cache_key(self, request):
        """ Returns the key used to store attachment files states into the file based cache. """
        return 'attachments_{}'.format(request.session.session_key)

    def get_post_form(self, form_class):
        """ Returns an instance of the post form to be used in the view. """
        return form_class(**self.get_post_form_kwargs())

    def get_post_form_class(self):
        """ Returns the post form class to use for instantiating the form. """
        return self.post_form_class

    def get_post_form_kwargs(self):
        """ Returns the keyword arguments for instantiating the post form. """
        kwargs = {
            'user': self.request.user,
            'forum': self.get_forum(),
            'topic': self.get_topic(),
        }

        post = self.get_post()
        if post:
            kwargs.update({'instance': post})

        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        return kwargs

    def get_attachment_formset(self, formset_class):
        """ Returns an instance of the attachment formset to be used in the view. """
        if (
            self.request.forum_permission_handler.can_attach_files(
                self.get_forum(), self.request.user,
            )
        ):
            return formset_class(**self.get_attachment_formset_kwargs())

    def get_attachment_formset_class(self):
        """ Returns the attachment formset class to use to initialize the attachment formset. """
        return self.attachment_formset_class

    def get_attachment_formset_kwargs(self):
        """ Returns the keyword arguments for instantiating the attachment formset. """
        kwargs = {
            'prefix': 'attachment',
        }

        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        else:
            post = self.get_post()
            attachment_queryset = Attachment.objects.filter(post=post)
            kwargs.update({
                'queryset': attachment_queryset,
            })
        return kwargs

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = kwargs
        if 'view' not in context:
            context['view'] = self

        # Insert the considered forum, topic and post into the context
        context['forum'] = self.get_forum()
        context['topic'] = self.get_topic()
        context['post'] = self.get_post()

        # Handles the preview of the attachments
        if context['attachment_formset']:
            if hasattr(self, 'attachment_preview') and self.attachment_preview:
                context['attachment_preview'] = self.attachment_preview
                attachments = []
                # Computes the list of the attachment file names that should be attached
                # to the forum post being created or updated
                for form in context['attachment_formset'].forms:
                    if (
                        form['DELETE'].value() or
                        (not form['file'].html_name in self.request._files and not form.instance.pk)
                    ):
                        continue
                    attachments.append(
                        (
                            form,
                            self.request._files[form['file'].html_name].name if not form.instance
                            else form.instance.filename
                        )
                    )
                context['attachment_file_previews'] = attachments

        return context

    def get_forum(self):
        """ Returns the considered forum. """
        pk = self.kwargs.get(self.forum_pk_url_kwarg, None)
        if not pk:  # pragma: no cover
            # This should never happen
            return
        if not hasattr(self, '_forum'):
            self._forum = get_object_or_404(Forum, pk=pk)
        return self._forum

    def get_topic(self):
        """ Returns the considered topic if applicable. """
        pk = self.kwargs.get(self.topic_pk_url_kwarg, None)
        if not pk:
            return
        if not hasattr(self, '_topic'):
            self._topic = get_object_or_404(Topic, pk=pk)
        return self._topic

    def get_post(self):
        """ Returns the considered post if applicable. """
        pk = self.kwargs.get(self.post_pk_url_kwarg, None)
        if not pk:
            return
        if not hasattr(self, '_forum_post'):
            self._forum_post = get_object_or_404(Post, pk=pk)
        return self._forum_post

    def form_valid(self, post_form, attachment_formset, **kwargs):
        """ Processes valid forms.

        Called if all forms are valid. Creates a Post instance along with associated attachments if
        required and then redirects to a success page.

        """
        save_attachment_formset = attachment_formset is not None \
            and not self.preview

        if self.preview:
            return self.render_to_response(
                self.get_context_data(
                    preview=True, post_form=post_form, attachment_formset=attachment_formset,
                    **kwargs
                ),
            )

        # This is not a preview ; the object is going to be saved
        self.forum_post = post_form.save()

        if save_attachment_formset:
            attachment_formset.post = self.forum_post
            attachment_formset.save()

        messages.success(self.request, self.success_message)
        if not self.forum_post.approved:
            messages.warning(self.request, self.approval_required_message)

        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, post_form, attachment_formset, **kwargs):
        """ Processes invalid forms.

        Called if one of the forms is invalid. Re-renders the context data with the data-filled
        forms and errors.

        """
        if (
            attachment_formset and
            not attachment_formset.is_valid() and
            len(attachment_formset.errors)
        ):
            messages.error(self.request, self.attachment_formset_general_error_message)

        return self.render_to_response(
            self.get_context_data(
                post_form=post_form, attachment_formset=attachment_formset, **kwargs
            ),
        )


class BaseTopicFormView(BasePostFormView):
    """ A base view for handling topic forms. """

    poll_option_formset_class = TopicPollOptionFormset
    poll_option_formset_general_error_message = _(
        'There are some errors in the poll options you submitted.'
    )
    # A specific form class is used here in order to fill the Topic-related date in addition to the
    # ones related to Post instances.
    post_form_class = TopicForm

    def get(self, request, *args, **kwargs):
        """ Handles GET requests. """
        self.init_attachment_cache()

        # Initializes the forms
        post_form_class = self.get_post_form_class()
        post_form = self.get_post_form(post_form_class)
        attachment_formset_class = self.get_attachment_formset_class()
        attachment_formset = self.get_attachment_formset(attachment_formset_class)
        poll_option_formset_class = self.get_poll_option_formset_class()
        poll_option_formset = self.get_poll_option_formset(poll_option_formset_class)

        return self.render_to_response(
            self.get_context_data(
                post_form=post_form, attachment_formset=attachment_formset,
                poll_option_formset=poll_option_formset,
            ),
        )

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        self.init_attachment_cache()

        # Stores a boolean indicating if we are considering a preview
        self.preview = 'preview' in self.request.POST

        # Initializes the forms
        post_form_class = self.get_post_form_class()
        post_form = self.get_post_form(post_form_class)
        attachment_formset_class = self.get_attachment_formset_class()
        attachment_formset = self.get_attachment_formset(attachment_formset_class)
        poll_option_formset_class = self.get_poll_option_formset_class()
        poll_option_formset = self.get_poll_option_formset(poll_option_formset_class)

        post_form_valid = post_form.is_valid()
        attachment_formset_valid = (
            attachment_formset.is_valid() if attachment_formset else None
        )
        poll_option_formset_valid = (
            poll_option_formset.is_valid()
            if poll_option_formset and len(post_form.cleaned_data['poll_question']) else None
        )

        self.attachment_preview = self.preview if attachment_formset_valid else None
        self.poll_preview = self.preview if poll_option_formset_valid else None

        poll_options_validated = poll_option_formset_valid is not None
        if (
            post_form_valid and
            attachment_formset_valid is not False and
            poll_option_formset_valid is not False
        ):
            return self.form_valid(
                post_form, attachment_formset, poll_option_formset,
                poll_options_validated=poll_options_validated,
            )
        else:
            return self.form_invalid(
                post_form, attachment_formset, poll_option_formset,
                poll_options_validated=poll_options_validated,
            )

    def get_poll_option_formset(self, formset_class):
        """ Returns an instance of the poll option formset to be used in the view. """
        if self.request.forum_permission_handler.can_create_polls(
            self.get_forum(), self.request.user,
        ):
            return formset_class(**self.get_poll_option_formset_kwargs())

    def get_poll_option_formset_class(self):
        """ Returns the poll option formset class to use to initialize the poll option formset. """
        return self.poll_option_formset_class

    def get_poll_option_formset_kwargs(self):
        """ Returns the keyword arguments for instantiating the poll option formset. """
        kwargs = {
            'prefix': 'poll',
        }

        if self.request.method in ('POST', 'PUT'):
            kwargs.update({
                'data': self.request.POST,
                'files': self.request.FILES,
            })
        else:
            topic = self.get_topic()
            poll_option_queryset = TopicPollOption.objects.filter(poll__topic=topic)
            kwargs.update({
                'queryset': poll_option_queryset,
            })
        return kwargs

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)

        # Handles the preview of the poll
        if context['poll_option_formset']:
            if hasattr(self, 'poll_preview') and self.poll_preview:
                context['poll_preview'] = self.poll_preview
                context['poll_options_previews'] = filter(
                    lambda f: f['text'].value() and not f['DELETE'].value(),
                    context['poll_option_formset'].forms,
                )

        return context

    def form_valid(self, post_form, attachment_formset, poll_option_formset, **kwargs):
        """ Processes valid forms. """
        save_poll_option_formset = poll_option_formset is not None \
            and not self.preview

        valid = super().form_valid(
            post_form, attachment_formset, poll_option_formset=poll_option_formset, **kwargs)

        if save_poll_option_formset:
            poll_option_formset.topic = self.forum_post.topic
            poll_option_formset.save(
                poll_question=post_form.cleaned_data.pop('poll_question', None),
                poll_max_options=post_form.cleaned_data.pop('poll_max_options', None),
                poll_duration=post_form.cleaned_data.pop('poll_duration', None),
                poll_user_changes=post_form.cleaned_data.pop('poll_user_changes', None),
            )

        return valid

    def form_invalid(self, post_form, attachment_formset, poll_option_formset, **kwargs):
        """ Processes invalid forms. """
        poll_errors = [k for k in post_form.errors.keys() if k.startswith('poll_')]
        if (
            poll_errors or
            (
                poll_option_formset and
                not poll_option_formset.is_valid() and
                len(post_form.cleaned_data['poll_question'])
            )
        ):
            messages.error(self.request, self.poll_option_formset_general_error_message)

        return super().form_invalid(
            post_form, attachment_formset, poll_option_formset=poll_option_formset, **kwargs)


class PostFormView(SingleObjectMixin, BasePostFormView):
    """ A base view for manipulating post forms. """

    forum_pk_url_kwarg = 'forum_pk'
    post_pk_url_kwarg = 'pk'
    topic_pk_url_kwarg = 'topic_pk'


class TopicFormView(SingleObjectMixin, BaseTopicFormView):
    """ A base view for manipulating topic forms. """

    forum_pk_url_kwarg = 'forum_pk'
    topic_pk_url_kwarg = 'pk'


class TopicCreateView(PermissionRequiredMixin, TopicFormView):
    """ Allows users to create forum topics. """

    model = Topic
    permission_required = ['can_start_new_topics', ]
    template_name = 'forum_conversation/topic_create.html'

    def get(self, request, *args, **kwargs):
        """ Handles GET requests. """
        self.object = None
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        self.object = None
        return super().post(request, *args, **kwargs)

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_forum()

    def get_success_url(self):
        """ Returns the URL to redirect the user to upon valid form processing. """
        if not self.forum_post.approved:
            return reverse(
                'forum:forum',
                kwargs={
                    'slug': self.forum_post.topic.forum.slug,
                    'pk': self.forum_post.topic.forum.pk,
                },
            )
        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.forum_post.topic.forum.slug,
                'forum_pk': self.forum_post.topic.forum.pk,
                'slug': self.forum_post.topic.slug,
                'pk': self.forum_post.topic.pk,
            },
        )


class TopicUpdateView(PermissionRequiredMixin, TopicFormView):
    """ Allows users to update forum topics. """

    model = Topic
    success_message = _('This message has been edited successfully.')
    template_name = 'forum_conversation/topic_update.html'

    def get(self, request, *args, **kwargs):
        """ Handles GET requests. """
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_topic().first_post

    def get_object(self, queryset=None):
        """ Returns the considered object. """
        return self.get_topic()

    def get_post(self):
        """ Returns the considered post if applicable. """
        return self.get_topic().first_post

    def get_success_url(self):
        """ Returns the URL to redirect the user to upon valid form processing. """
        return reverse(
            'forum_conversation:topic', kwargs={
                'forum_slug': self.forum_post.topic.forum.slug,
                'forum_pk': self.forum_post.topic.forum.pk,
                'slug': self.forum_post.topic.slug,
                'pk': self.forum_post.topic.pk,
            },
        )

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_edit_post(obj, user)


class PostCreateView(PermissionRequiredMixin, PostFormView):
    """ Allows users to create forum posts. """

    model = Post
    template_name = 'forum_conversation/post_create.html'

    def get(self, request, *args, **kwargs):
        """ Handles GET requests. """
        self.object = None
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        self.object = None
        return super().post(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)
        topic = self.get_topic()

        # Add the previous posts to the context
        previous_posts = (
            topic.posts.filter(approved=True)
            .select_related('poster', 'updated_by')
            .prefetch_related('attachments', 'poster__forum_profile')
            .order_by('-created')
        )
        previous_posts = previous_posts[:machina_settings.TOPIC_REVIEW_POSTS_NUMBER]
        context['previous_posts'] = previous_posts

        return context

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_topic()

    def get_success_url(self):
        """ Returns the URL to redirect the user to upon valid form processing. """
        return '{0}?post={1}#{1}'.format(
            reverse(
                'forum_conversation:topic',
                kwargs={
                    'forum_slug': self.forum_post.topic.forum.slug,
                    'forum_pk': self.forum_post.topic.forum.pk,
                    'slug': self.forum_post.topic.slug,
                    'pk': self.forum_post.topic.pk,
                },
            ),
            self.forum_post.pk,
        )

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_add_post(obj, user)


class PostUpdateView(PermissionRequiredMixin, PostFormView):
    """ Allows users to update forum topics. """

    model = Post
    success_message = _('This message has been edited successfully.')
    template_name = 'forum_conversation/post_update.html'

    def get(self, request, *args, **kwargs):
        """ Handles GET requests. """
        self.object = self.get_object()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """ Handles POST requests. """
        self.object = self.get_object()
        return super().post(request, *args, **kwargs)

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_post()

    def get_object(self, queryset=None):
        """ Returns the considered object. """
        return self.get_post()

    def get_success_url(self):
        """ Returns the URL to redirect the user to upon valid form processing. """
        return '{0}?post={1}#{1}'.format(
            reverse(
                'forum_conversation:topic',
                kwargs={
                    'forum_slug': self.forum_post.topic.forum.slug,
                    'forum_pk': self.forum_post.topic.forum.pk,
                    'slug': self.forum_post.topic.slug,
                    'pk': self.forum_post.topic.pk,
                },
            ),
            self.forum_post.pk,
        )

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_edit_post(obj, user)


class PostDeleteView(PermissionRequiredMixin, DeleteView):
    """ Allows users to delete forum topics. """

    model = Post
    success_message = _('This message has been deleted successfully.')
    template_name = 'forum_conversation/post_delete.html'

    def get_context_data(self, **kwargs):
        """ Returns the context data to provide to the template. """
        context = super().get_context_data(**kwargs)

        # Append the topic and the forum associated with the post being deleted
        # to the context
        post = self.get_object()
        context['topic'] = post.topic
        context['forum'] = post.topic.forum

        return context

    def get_controlled_object(self):
        """ Returns the controlled object. """
        return self.get_object()

    def get_success_url(self):
        """ Returns the URL to redirect the user to upon valid form processing. """
        messages.success(self.request, self.success_message)

        if self.object.is_topic_head and self.object.is_topic_tail:
            return reverse(
                'forum:forum',
                kwargs={
                    'slug': self.object.topic.forum.slug, 'pk': self.object.topic.forum.pk,
                },
            )

        return reverse(
            'forum_conversation:topic',
            kwargs={
                'forum_slug': self.object.topic.forum.slug,
                'forum_pk': self.object.topic.forum.pk,
                'slug': self.object.topic.slug,
                'pk': self.object.topic.pk,
            },
        )

    def perform_permissions_check(self, user, obj, perms):
        """ Performs the permission check. """
        return self.request.forum_permission_handler.can_delete_post(obj, user)



from django.core import validators
from django.core.exceptions import ImproperlyConfigured
from django.utils.deconstruct import deconstructible
from django.utils.encoding import smart_str
from django.utils.translation import gettext_lazy as _

from machina.conf import settings as machina_settings


class NullableMaxLengthValidator(validators.MaxLengthValidator):
    """ Provides a way to not validate an input if the max length is None """

    def __call__(self, value):
        if self.limit_value is None:
            # If the limit value is None, this means that there is no
            # limit value at all. The default validation process is not
            # performed.
            return
        super().__call__(value)


def _get_markup_maxlength_validator():
    try:
        dotted_path = machina_settings.MARKUP_MAX_LENGTH_VALIDATOR
        module, validator = dotted_path.rsplit('.', 1)
        module, validator = smart_str(module), smart_str(validator)
        validator = getattr(__import__(module, {}, {}, [validator]), validator)
        return validator
    except ImportError as e:
        raise ImproperlyConfigured(
            _('Could not import MACHINA MARKUP_MAX_LENGTH_VALIDATOR {}: {}').format(
                machina_settings.MARKUP_MAX_LENGTH_VALIDATOR, e
            )
        )
    except AttributeError:
        raise ImproperlyConfigured(_('MACHINA_MARKUP_MAX_LENGTH_VALIDATOR setting is required'))


ConfiguredMarkupMaxLengthValidator = _get_markup_maxlength_validator()


@deconstructible
class MarkupMaxLengthValidator:
    """
    Validates the max length of an input written in the markup language defined in the settings.

    This is a proxy that use the implementation defined in MACHINA_MARKUP_MAX_LENGTH_VALIDATOR
    setting.
    """

    def __init__(self, limit_value):
        self.validator = ConfiguredMarkupMaxLengthValidator(limit_value)

    def __call__(self, value):
        self.validator(value)



"""
    Forum feeds URLs
    ================

    This module defines URL patterns associated with the django-machina's ``forum_feeds``
    application.

"""

from django.urls import path

from machina.core.loading import get_class
from machina.core.urls import URLPatternsFactory


class ForumFeedsURLPatternsFactory(URLPatternsFactory):
    """ Allows to generate the URL patterns of the ``forum_feeds`` application. """

    app_namespace = 'forum_feeds'

    latest_topics_feed = get_class('forum_feeds.feeds', 'LastTopicsFeed')

    def get_urlpatterns(self):
        """ Returns the URL patterns managed by the considered factory / application. """
        return [
            path('topics/', self.latest_topics_feed(), name='latest_topics'),
            path(
                'forum/<str:forum_slug>-<int:forum_pk>/topics/',
                self.latest_topics_feed(),
                name='forum_latest_topics',
            ),
            path(
                'forum/<str:forum_slug>-<int:forum_pk>/topics/all/',
                self.latest_topics_feed(),
                {'descendants': True},
                name='forum_latest_topics_with_descendants',
            ),
        ]


urlpatterns_factory = ForumFeedsURLPatternsFactory()
